from typing import Any, Generic, List, Literal, Optional, Type, Union, cast, overload

from weaviate.collections.classes.filters import _Filters
from weaviate.collections.classes.grpc import METADATA, PROPERTIES, REFERENCES, Sorting
from weaviate.collections.classes.internal import (
    CrossReferences,
    GenerativeReturn,
    GenerativeReturnType,
    ReturnProperties,
    ReturnReferences,
    _Generative,
    _GenerativeConfigRuntime,
    _GroupedTask,
    _QueryOptions,
    _SinglePrompt,
)
from weaviate.collections.classes.types import (
    Properties,
    References,
    TProperties,
    TReferences,
)
from weaviate.collections.queries.base_executor import _BaseExecutor
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionType
from weaviate.proto.v1 import search_get_pb2
from weaviate.types import INCLUDE_VECTOR, UUID


class _FetchObjectsGenerateExecutor(
    Generic[ConnectionType, Properties, References], _BaseExecutor[ConnectionType]
):
    @overload
    def fetch_objects(
        self,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Literal[None] = None,
    ) -> executor.Result[GenerativeReturn[Properties, References]]: ...

    @overload
    def fetch_objects(
        self,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: REFERENCES,
    ) -> executor.Result[GenerativeReturn[Properties, CrossReferences]]: ...

    @overload
    def fetch_objects(
        self,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Type[TReferences],
    ) -> executor.Result[GenerativeReturn[Properties, TReferences]]: ...

    @overload
    def fetch_objects(
        self,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: Literal[None] = None,
    ) -> executor.Result[GenerativeReturn[TProperties, References]]: ...

    @overload
    def fetch_objects(
        self,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: REFERENCES,
    ) -> executor.Result[GenerativeReturn[TProperties, CrossReferences]]: ...

    @overload
    def fetch_objects(
        self,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: Type[TReferences],
    ) -> executor.Result[GenerativeReturn[TProperties, TReferences]]: ...

    @overload
    def fetch_objects(
        self,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
        return_references: Optional[ReturnReferences[TReferences]] = None,
    ) -> executor.Result[
        GenerativeReturnType[Properties, References, TProperties, TReferences]
    ]: ...

    def fetch_objects(
        self,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
        return_references: Optional[ReturnReferences[TReferences]] = None,
    ) -> executor.Result[GenerativeReturnType[Properties, References, TProperties, TReferences]]:
        """Perform retrieval-augmented generation (RaG) on the results of a simple get query of objects in this collection.

        Args:
            single_prompt: The prompt to use for RaG on each object individually.
            grouped_task: The prompt to use for RaG on the entire result set.
            grouped_properties: The properties to use in the RaG on the entire result set.
            limit: The maximum number of results to return. If not specified, the default limit specified by Weaviate is returned.
            offset: The offset to start from. If not specified, the retrieval begins from the first object in Weaviate.
            after: The UUID of the object to start from. If not specified, the retrieval begins from the first object in Weaviate.
            filters: The filters to apply to the retrieval.
            sort: The sorting to apply to the retrieval.
            include_vector: Whether to include the vector in the results. If not specified, this is set to False.
            return_metadata: The metadata to return for each object, defaults to `None`.
            return_properties: The properties to return for each object.
            return_references: The references to return for each object.

        NOTE:
            - If `return_properties` is not provided then all properties are returned except for blob properties.
            - If `return_metadata` is not provided then no metadata is provided. Use MetadataQuery.full() to retrieve all metadata.
            - If `return_references` is not provided then no references are provided.

        Returns:
            A `_GenerativeNearMediaReturn` object that includes the searched objects with per-object generated results and group generated results.

        Raises:
            weaviate.exceptions.WeaviateGRPCQueryError: If the network connection to Weaviate fails.
        """

        def resp(
            res: search_get_pb2.SearchReply,
        ) -> GenerativeReturnType[Properties, References, TProperties, TReferences]:
            return cast(
                Any,
                self._result_to_generative_query_return(
                    res,
                    _QueryOptions.from_input(
                        return_metadata,
                        return_properties,
                        include_vector,
                        self._references,
                        return_references,
                    ),
                ),
            )

        request = self._query.get(
            limit=limit,
            offset=offset,
            after=after,
            filters=filters,
            sort=sort,
            return_metadata=self._parse_return_metadata(return_metadata, include_vector),
            return_properties=self._parse_return_properties(return_properties),
            return_references=self._parse_return_references(return_references),
            generative=_Generative(
                single=single_prompt,
                grouped=grouped_task,
                grouped_properties=grouped_properties,
                generative_provider=generative_provider,
            ),
        )
        return executor.execute(
            response_callback=resp,
            method=self._connection.grpc_search,
            request=request,
        )



from typing import List, cast

from weaviate import (
    WeaviateAsyncClient,
    WeaviateClient,
    connect_to_local,
    use_async_with_local,
)
from weaviate.classes.config import DataType, Property


class SyncJourneys:
    def __init__(self, client: WeaviateClient) -> None:
        self.__client = client

    @classmethod
    def use(cls) -> "SyncJourneys":
        return cls(connect_to_local(port=8090, grpc_port=50061))

    def close(self) -> None:
        self.__client.close()

    def simple(self) -> List[dict]:
        name = "FastAPISyncTestingCollection"
        if self.__client.collections.exists(name):
            self.__client.collections.delete(name)
        collection = self.__client.collections.create(
            name=name,
            properties=[
                Property(name="name", data_type=DataType.TEXT),
                Property(name="age", data_type=DataType.INT),
            ],
        )
        with collection.batch.dynamic() as batch:
            for i in range(1000):
                batch.add_object({"name": f"Person {i}", "age": i})
        res = collection.query.fetch_objects(limit=100)
        self.__client.collections.delete(name)
        return [cast(dict, obj.properties) for obj in res.objects]


class AsyncJourneys:
    def __init__(self, client: WeaviateAsyncClient) -> None:
        self.__client = client

    @classmethod
    async def use(cls) -> "AsyncJourneys":
        client = use_async_with_local(port=8090, grpc_port=50061)
        await client.connect()
        return cls(client)

    async def close(self) -> None:
        await self.__client.close()

    async def simple(self) -> List[dict]:
        name = "FastAPIAsyncTestingCollection"
        if await self.__client.collections.exists(name):
            await self.__client.collections.delete(name)
        collection = await self.__client.collections.create(
            name=name,
            properties=[
                Property(name="name", data_type=DataType.TEXT),
                Property(name="age", data_type=DataType.INT),
            ],
        )
        await collection.data.insert_many([{"name": f"Person {i}", "age": i} for i in range(100)])
        res = await collection.query.fetch_objects(limit=1000)
        await self.__client.collections.delete(name)
        return [cast(dict, obj.properties) for obj in res.objects]



from weaviate.cluster.base import _ClusterExecutor
from weaviate.cluster.replicate import _ReplicateAsync
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionAsync


@executor.wrap("async")
class _ClusterAsync(_ClusterExecutor[ConnectionAsync]):
    def __init__(self, connection: ConnectionAsync):
        super().__init__(connection)
        self.replications = _ReplicateAsync(connection)



from typing import Generic, List, Literal, Optional, Type, Union, overload

from weaviate.collections.classes.filters import _Filters
from weaviate.collections.classes.grpc import (
    METADATA,
    PROPERTIES,
    REFERENCES,
    GroupBy,
    Rerank,
    TargetVectorJoinType,
)
from weaviate.collections.classes.internal import (
    CrossReferences,
    GenerativeGroupByReturn,
    GenerativeReturn,
    GenerativeSearchReturnType,
    ReturnProperties,
    ReturnReferences,
    _GenerativeConfigRuntime,
    _GroupedTask,
    _SinglePrompt,
)
from weaviate.collections.classes.types import Properties, References, TProperties, TReferences
from weaviate.connect.v4 import ConnectionSync
from weaviate.types import INCLUDE_VECTOR, NUMBER, UUID

from .executor import _NearObjectGenerateExecutor

class _NearObjectGenerate(
    Generic[Properties, References],
    _NearObjectGenerateExecutor[ConnectionSync, Properties, References],
):
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Literal[None] = None,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Literal[None] = None,
    ) -> GenerativeReturn[Properties, References]: ...
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Literal[None] = None,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: REFERENCES,
    ) -> GenerativeReturn[Properties, CrossReferences]: ...
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Literal[None] = None,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Type[TReferences],
    ) -> GenerativeReturn[Properties, TReferences]: ...
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Literal[None] = None,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: Literal[None] = None,
    ) -> GenerativeReturn[TProperties, References]: ...
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Literal[None] = None,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: REFERENCES,
    ) -> GenerativeReturn[TProperties, CrossReferences]: ...
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Literal[None] = None,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: Type[TReferences],
    ) -> GenerativeReturn[TProperties, TReferences]: ...
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: GroupBy,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Literal[None] = None,
    ) -> GenerativeGroupByReturn[Properties, References]: ...
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: GroupBy,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: REFERENCES,
    ) -> GenerativeGroupByReturn[Properties, CrossReferences]: ...
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: GroupBy,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Type[TReferences],
    ) -> GenerativeGroupByReturn[Properties, TReferences]: ...
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: GroupBy,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: Literal[None] = None,
    ) -> GenerativeGroupByReturn[TProperties, References]: ...
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: GroupBy,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: REFERENCES,
    ) -> GenerativeGroupByReturn[TProperties, CrossReferences]: ...
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: GroupBy,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: Type[TReferences],
    ) -> GenerativeGroupByReturn[TProperties, TReferences]: ...
    @overload
    def near_object(
        self,
        near_object: UUID,
        *,
        single_prompt: Union[str, _SinglePrompt, None] = None,
        grouped_task: Union[str, _GroupedTask, None] = None,
        grouped_properties: Optional[List[str]] = None,
        generative_provider: Optional[_GenerativeConfigRuntime] = None,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[GroupBy] = None,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
        return_references: Optional[ReturnReferences[TReferences]] = None,
    ) -> GenerativeSearchReturnType[Properties, References, TProperties, TReferences]: ...



from weaviate.collections.tenants.executor import _TenantsExecutor
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionSync


@executor.wrap("sync")
class _Tenants(_TenantsExecutor[ConnectionSync]):
    pass



from weaviate.collections.aggregations.near_text.executor import _NearTextExecutor
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionSync


@executor.wrap("sync")
class _NearText(_NearTextExecutor[ConnectionSync]):
    pass



from typing import Dict, Generic, List, Literal, Union, overload

from httpx import Response

from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionType, _ExpectedStatusCodes
from weaviate.rbac.models import (
    Role,
    RoleBase,
)
from weaviate.users.users import (
    USER_TYPE,
    USER_TYPE_OIDC,
)
from weaviate.util import escape_string


class _BaseExecutor(Generic[ConnectionType]):
    def __init__(self, connection: ConnectionType):
        self._connection = connection

    def _get_roles_of_group(
        self,
        group_id: str,
        group_type: USER_TYPE,
        include_permissions: bool,
    ) -> executor.Result[Union[Dict[str, Role], Dict[str, RoleBase]]]:
        path = f"/authz/groups/{escape_string(group_id)}/roles/{group_type}"

        def resp(res: Response) -> Union[Dict[str, Role], Dict[str, RoleBase]]:
            roles = res.json()
            if include_permissions:
                return {role["name"]: Role._from_weaviate_role(role) for role in roles}
            return {role["name"]: RoleBase(role["name"]) for role in roles}

        return executor.execute(
            response_callback=resp,
            method=self._connection.get,
            path=path,
            params={"includeFullRoles": include_permissions},
            error_msg=f"Could not get roles of group {group_id}",
            status_codes=_ExpectedStatusCodes(ok_in=[200], error="Get roles of group"),
        )

    def _assign_roles_to_group(
        self,
        roles: List[str],
        group_id: str,
        group_type: USER_TYPE,
    ) -> executor.Result[None]:
        path = f"/authz/groups/{escape_string(group_id)}/assign"

        def resp(res: Response) -> None:
            pass

        return executor.execute(
            response_callback=resp,
            method=self._connection.post,
            path=path,
            weaviate_object={"roles": roles, "groupType": group_type},
            error_msg=f"Could not assign roles {roles} to group {group_id}",
            status_codes=_ExpectedStatusCodes(ok_in=[200], error="Assign group to roles"),
        )

    def _revoke_roles_from_group(
        self,
        roles: Union[str, List[str]],
        group_id: str,
        group_type: USER_TYPE,
    ) -> executor.Result[None]:
        path = f"/authz/groups/{escape_string(group_id)}/revoke"

        def resp(res: Response) -> None:
            pass

        return executor.execute(
            response_callback=resp,
            method=self._connection.post,
            path=path,
            weaviate_object={"roles": roles, "groupType": group_type},
            error_msg=f"Could not revoke roles {roles} from group {group_id}",
            status_codes=_ExpectedStatusCodes(ok_in=[200], error="Revoke group from user"),
        )

    def _get_known_group_names(self, group_type: USER_TYPE) -> executor.Result[List[str]]:
        path = f"/authz/groups/{group_type}"

        def resp(res: Response) -> List[str]:
            groups = res.json()
            assert isinstance(groups, list), "Expected a list of group names"
            return groups

        return executor.execute(
            response_callback=resp,
            method=self._connection.get,
            path=path,
            error_msg=f"Could not get known groups for group type {group_type}",
            status_codes=_ExpectedStatusCodes(ok_in=[200], error="Get known groups"),
        )


class _GroupsOIDCExecutor(Generic[ConnectionType], _BaseExecutor[ConnectionType]):
    @overload
    def get_assigned_roles(
        self, *, group_id: str, include_permissions: Literal[False] = False
    ) -> executor.Result[Dict[str, RoleBase]]: ...

    @overload
    def get_assigned_roles(
        self, *, group_id: str, include_permissions: Literal[True]
    ) -> executor.Result[Dict[str, Role]]: ...

    @overload
    def get_assigned_roles(
        self,
        *,
        group_id: str,
        include_permissions: bool = False,
    ) -> executor.Result[Union[Dict[str, Role], Dict[str, RoleBase]]]: ...

    def get_assigned_roles(
        self,
        *,
        group_id: str,
        include_permissions: bool = False,
    ) -> executor.Result[Union[Dict[str, Role], Dict[str, RoleBase]]]:
        """Get the roles assigned to a group specific to the configured OIDC's dynamic auth functionality.

        Args:
            group_id: The group ID to get the roles for.

        Returns:
            A dictionary with role names as keys and the `Role` objects as values.
        """
        return self._get_roles_of_group(
            group_id,
            USER_TYPE_OIDC,
            include_permissions,
        )

    def assign_roles(
        self,
        *,
        group_id: str,
        role_names: Union[str, List[str]],
    ) -> executor.Result[None]:
        """Assign roles to a group specific to the configured OIDC's dynamic auth functionality.

        Args:
            role_names: The names of the roles to assign to the group.
            group_id: The group to assign the roles to.
        """
        return self._assign_roles_to_group(
            [role_names] if isinstance(role_names, str) else role_names,
            group_id,
            USER_TYPE_OIDC,
        )

    def revoke_roles(
        self,
        *,
        group_id: str,
        role_names: Union[str, List[str]],
    ) -> executor.Result[None]:
        """Revoke roles from a group specific to the configured OIDC's dynamic auth functionality.

        Args:
            role_names: The names of the roles to revoke from the group.
            group_id: The group to revoke the roles from.
        """
        return self._revoke_roles_from_group(
            [role_names] if isinstance(role_names, str) else role_names,
            group_id,
            USER_TYPE_OIDC,
        )

    def get_known_group_names(self) -> executor.Result[List[str]]:
        """Get the known group names specific to the configured OIDC's dynamic auth functionality.

        Returns:
            A list of known group names.
        """
        return self._get_known_group_names(
            USER_TYPE_OIDC,
        )



import warnings


warnings.filterwarnings(
    "ignore",
    r"(?=.*5\.27\.\d+)(?=.*5\.28\.\d+)(?=.*obsolete)",
    UserWarning,
    "google.protobuf.runtime_version",
)
# ref: https://github.com/grpc/grpc/issues/37609 and https://github.com/protocolbuffers/protobuf/pull/17241

from packaging import version

from importlib.metadata import version as metadata_version

from weaviate.exceptions import WeaviateProtobufIncompatibility

def get_version(pkg: str)-> version.Version:
    return version.parse(metadata_version(pkg))

pb_version, grpc_version = get_version("protobuf"), get_version("grpcio")
if pb_version >= version.parse("6.30.0"):
    if grpc_version < version.parse("1.72.0"):
        raise WeaviateProtobufIncompatibility(pb_version, grpc_version)
    from weaviate.proto.v1.v6300.v1 import weaviate_pb2_grpc, aggregate_pb2, base_pb2, base_search_pb2, batch_delete_pb2, batch_pb2, generative_pb2, health_weaviate_pb2, health_weaviate_pb2_grpc, properties_pb2, search_get_pb2, tenants_pb2
elif pb_version >= version.parse("5.26.1"):
    if grpc_version < version.parse("1.63.0"):
        raise WeaviateProtobufIncompatibility(pb_version, grpc_version)
    from weaviate.proto.v1.v5261.v1 import weaviate_pb2_grpc, aggregate_pb2, base_pb2, base_search_pb2, batch_delete_pb2, batch_pb2, generative_pb2, health_weaviate_pb2, health_weaviate_pb2_grpc, properties_pb2, search_get_pb2, tenants_pb2
elif pb_version >= version.parse("4.21.6"):
    from weaviate.proto.v1.v4216.v1 import weaviate_pb2_grpc, aggregate_pb2, base_pb2, base_search_pb2, batch_delete_pb2, batch_pb2, generative_pb2, health_weaviate_pb2, health_weaviate_pb2_grpc, properties_pb2, search_get_pb2, tenants_pb2
else:
    raise RuntimeError(f"Unsupported protobuf version: {pb_version}. Only versions 4.21.6+ are supported.")

__all__ = [
    "aggregate_pb2", "base_pb2", "base_search_pb2", "batch_delete_pb2", "batch_pb2", "generative_pb2", "health_weaviate_pb2", "health_weaviate_pb2_grpc", "properties_pb2", "search_get_pb2", "tenants_pb2", "weaviate_pb2_grpc"
]


import asyncio
from math import ceil
from typing import Any, Dict, Generic, List, Optional, Sequence, Union

from httpx import Response

from weaviate.collections.classes.tenants import (
    Tenant,
    TenantActivityStatus,
    TenantCreate,
    TenantCreateActivityStatus,
    TenantOutput,
    TenantUpdate,
    TenantUpdateActivityStatus,
)
from weaviate.collections.grpc.tenants import _TenantsGRPC
from weaviate.collections.tenants.types import (
    TenantCreateInputType,
    TenantInputType,
    TenantOutputType,
    TenantUpdateInputType,
)
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionAsync, ConnectionType, _ExpectedStatusCodes
from weaviate.exceptions import WeaviateInvalidInputError
from weaviate.proto.v1 import tenants_pb2
from weaviate.validator import _validate_input, _ValidateArgument

UPDATE_TENANT_BATCH_SIZE = 100


class _TenantsExecutor(Generic[ConnectionType]):
    def __init__(
        self,
        connection: ConnectionType,
        name: str,
        validate_arguments: bool = True,
    ) -> None:
        self._connection = connection
        self._name = name
        self._grpc = _TenantsGRPC(
            weaviate_version=connection._weaviate_version,
            name=name,
        )
        self._validate_arguments = validate_arguments

    def create(
        self,
        tenants: Union[TenantCreateInputType, Sequence[TenantCreateInputType]],
    ) -> executor.Result[None]:
        """Create the specified tenants for this collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Args:
            tenants: A tenant name, `wvc.config.tenants.Tenant`, `wvc.config.tenants.TenantCreateInput` object, or a list of tenants names
                and/or `wvc.config.tenants.Tenant` objects to add to the given collection.
                If a string is provided, the tenant will be added with the default activity status of `HOT`.

        Raises:
            weaviate.exceptions.WeaviateConnectionError: If the network connection to Weaviate fails.
            weaviate.exceptions.UnexpectedStatusCodeError: If Weaviate reports a non-OK status.
            weaviate.exceptions.WeaviateInvalidInputError: If `tenants` is not a list of `wvc.Tenant` objects.
        """
        if self._validate_arguments:
            _validate_input(
                [
                    _ValidateArgument(
                        expected=[
                            str,
                            Tenant,
                            TenantCreate,
                            Sequence[Union[str, Tenant, TenantCreate]],
                        ],
                        name="tenants",
                        value=tenants,
                    )
                ]
            )
        path = "/schema/" + self._name + "/tenants"

        def resp(res: Response) -> None:
            return None

        return executor.execute(
            response_callback=resp,
            method=self._connection.post,
            path=path,
            weaviate_object=self.__map_create_tenants(tenants),
            error_msg=f"Collection tenants may not have been added properly for {self._name}",
            status_codes=_ExpectedStatusCodes(
                ok_in=200, error=f"Add collection tenants for {self._name}"
            ),
        )

    def remove(
        self,
        tenants: Union[TenantInputType, Sequence[TenantInputType]],
    ) -> executor.Result[None]:
        """Remove the specified tenants from this collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Args:
            tenants: A tenant name, `wvc.config.tenants.Tenant` object, or a list of tenants names
                and/or `wvc.config.tenants.Tenant` objects to remove from the given class.

        Raises:
            weaviate.exceptions.WeaviateConnectionError: If the network connection to Weaviate fails.
            weaviate.exceptions.UnexpectedStatusCodeError: If Weaviate reports a non-OK status.
            weaviate.exceptions.WeaviateInvalidInputError: If `tenants` is not a list of strings.
        """
        if self._validate_arguments:
            _validate_input(
                [
                    _ValidateArgument(
                        expected=[
                            str,
                            Tenant,
                            Sequence[Union[str, Tenant]],
                        ],
                        name="tenants",
                        value=tenants,
                    )
                ]
            )

        tenant_names: List[str] = []
        if isinstance(tenants, str) or isinstance(tenants, Tenant):
            tenant_names = [tenants.name if isinstance(tenants, Tenant) else tenants]
        else:
            for tenant in tenants:
                tenant_names.append(tenant.name if isinstance(tenant, Tenant) else tenant)

        path = "/schema/" + self._name + "/tenants"

        def resp(res: Response) -> None:
            return None

        return executor.execute(
            response_callback=resp,
            method=self._connection.delete,
            path=path,
            weaviate_object=tenant_names,
            error_msg=f"Collection tenants may not have been deleted for {self._name}",
            status_codes=_ExpectedStatusCodes(
                ok_in=200, error=f"Delete collection tenants for {self._name}"
            ),
        )

    def __get_with_rest(
        self,
    ) -> executor.Result[Dict[str, TenantOutputType]]:
        path = "/schema/" + self._name + "/tenants"

        def resp(res: Response) -> Dict[str, TenantOutputType]:
            tenant_resp: List[Dict[str, Any]] = res.json()
            for tenant in tenant_resp:
                tenant["activityStatusInternal"] = tenant["activityStatus"]
                del tenant["activityStatus"]
            return {tenant["name"]: TenantOutput(**tenant) for tenant in tenant_resp}

        return executor.execute(
            response_callback=resp,
            method=self._connection.get,
            path=path,
            error_msg=f"Could not get collection tenants for {self._name}",
            status_codes=_ExpectedStatusCodes(
                ok_in=200, error=f"Get collection tenants for {self._name}"
            ),
        )

    def __get_with_grpc(
        self, *, tenants: Optional[Sequence[TenantInputType]] = None
    ) -> executor.Result[Dict[str, TenantOutputType]]:
        names = (
            [tenant.name if isinstance(tenant, Tenant) else tenant for tenant in tenants]
            if tenants is not None
            else tenants
        )
        request = tenants_pb2.TenantsGetRequest(
            collection=self._name,
            names=tenants_pb2.TenantNames(values=names) if names is not None else None,
        )

        def resp(res: tenants_pb2.TenantsGetReply) -> Dict[str, TenantOutputType]:
            return {
                tenant.name: TenantOutput(
                    name=tenant.name,
                    activity_status=self._grpc.map_activity_status(tenant.activity_status),
                )
                for tenant in res.tenants
            }

        return executor.execute(
            response_callback=resp,
            method=self._connection.grpc_tenants_get,
            request=request,
        )

    def __map_create_tenant(self, tenant: TenantCreateInputType) -> TenantCreate:
        if isinstance(tenant, str):
            return TenantCreate(name=tenant)
        if isinstance(tenant, Tenant):
            if tenant.activity_status not in [
                TenantActivityStatus.ACTIVE,
                TenantActivityStatus.INACTIVE,
            ]:
                raise WeaviateInvalidInputError(
                    f"Tenant activity status must be either 'ACTIVE' or 'INACTIVE'. Other statuses are read-only and cannot be set. Tenant: {tenant.name} had status: {tenant.activity_status}"
                )
            activity_status = TenantCreateActivityStatus(tenant.activity_status)
            return TenantCreate(name=tenant.name, activity_status=activity_status)
        return tenant

    def __map_update_tenant(self, tenant: TenantUpdateInputType) -> TenantUpdate:
        if isinstance(tenant, Tenant):
            if tenant.activity_status not in [
                TenantActivityStatus.ACTIVE,
                TenantActivityStatus.INACTIVE,
                TenantActivityStatus.OFFLOADED,
            ]:
                raise WeaviateInvalidInputError(
                    f"Tenant activity status must be one of 'ACTIVE', 'INACTIVE' or 'OFFLOADED'. Other statuses are read-only and cannot be set. Tenant: {tenant.name} had status: {tenant.activity_status}"
                )
            activity_status = TenantUpdateActivityStatus(tenant.activity_status)
            return TenantUpdate(name=tenant.name, activity_status=activity_status)
        return tenant

    def __map_create_tenants(
        self,
        tenants: Union[TenantCreateInputType, Sequence[TenantCreateInputType]],
    ) -> List[dict]:
        if (
            isinstance(tenants, str)
            or isinstance(tenants, Tenant)
            or isinstance(tenants, TenantCreate)
        ):
            return [self.__map_create_tenant(tenants).model_dump()]
        else:
            return [self.__map_create_tenant(tenant).model_dump() for tenant in tenants]

    def __map_update_tenants(
        self, tenants: Union[TenantUpdateInputType, Sequence[TenantUpdateInputType]]
    ) -> List[List[dict]]:
        if isinstance(tenants, Tenant) or isinstance(tenants, TenantUpdate):
            return [[self.__map_update_tenant(tenants).model_dump()]]
        else:
            batches = ceil(len(tenants) / UPDATE_TENANT_BATCH_SIZE)
            return [
                [
                    self.__map_update_tenant(tenants[i + b * UPDATE_TENANT_BATCH_SIZE]).model_dump()
                    for i in range(
                        min(
                            len(tenants) - b * UPDATE_TENANT_BATCH_SIZE,
                            UPDATE_TENANT_BATCH_SIZE,
                        )
                    )
                ]
                for b in range(batches)
            ]

    def get(self) -> executor.Result[Dict[str, TenantOutputType]]:
        """Return all tenants currently associated with this collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Raises:
            weaviate.exceptions.WeaviateConnectionError: If the network connection to Weaviate fails.
            weaviate.exceptions.UnexpectedStatusCodeError: If Weaviate reports a non-OK status.
        """

        def resp(res: Dict[str, TenantOutputType]) -> Dict[str, TenantOutputType]:
            return res

        return executor.execute(
            response_callback=resp,
            method=(
                self.__get_with_grpc
                if self._connection._weaviate_version.supports_tenants_get_grpc
                else self.__get_with_rest
            ),
        )

    def get_by_names(
        self, tenants: Sequence[TenantInputType]
    ) -> executor.Result[Dict[str, TenantOutputType]]:
        """Return named tenants currently associated with this collection in Weaviate.

        If the tenant does not exist, it will not be included in the response.
        If no names are provided, all tenants will be returned.
        The collection must have been created with multi-tenancy enabled.

        Args:
            tenants: Sequence of tenant names of wvc.tenants.Tenant objects to retrieve. To retrieve all tenants, use the `get` method.

        Raises:
            weaviate.exceptions.WeaviateConnectionError: If the network connection to Weaviate fails.
            weaviate.exceptions.UnexpectedStatusCodeError: If Weaviate reports a non-OK status.
        """
        self._connection._weaviate_version.check_is_at_least_1_25_0("The 'get_by_names' method")
        if self._validate_arguments:
            _validate_input(
                _ValidateArgument(
                    expected=[Sequence[Union[str, Tenant]]],
                    name="names",
                    value=tenants,
                )
            )
        return self.__get_with_grpc(tenants=tenants)

    def get_by_name(self, tenant: TenantInputType) -> executor.Result[Optional[TenantOutputType]]:
        """Return a specific tenant associated with this collection in Weaviate.

        If the tenant does not exist, `None` will be returned.

        The collection must have been created with multi-tenancy enabled.

        Args:
            tenant: The tenant to retrieve.

        Raises:
            weaviate.exceptions.WeaviateConnectionError: If the network connection to Weaviate fails.
            weaviate.exceptions.UnexpectedStatusCodeError: If Weaviate reports a non-OK status.
        """
        self._connection._weaviate_version.check_is_at_least_1_25_0("The 'get_by_name' method")
        if self._validate_arguments:
            _validate_input(
                _ValidateArgument(expected=[Union[str, Tenant]], name="tenant", value=tenant)
            )
        tenant_name = tenant.name if isinstance(tenant, Tenant) else tenant
        if self._connection._weaviate_version.is_lower_than(1, 28, 0):
            # For Weaviate versions < 1.28.0, we need to use the gRPC API
            # such versions don't have RBAC so the filtering issue doesn't exist therein
            def resp_grpc(res: Dict[str, TenantOutputType]) -> Optional[TenantOutputType]:
                return res.get(tenant_name)

            return executor.execute(
                response_callback=resp_grpc,
                method=self.__get_with_grpc,
                tenants=[tenant_name],
            )

        # For Weaviate versions >= 1.28.0, we need to use the REST API
        # as the gRPC API filters out tenants that are not accessible to the user
        # due to RBAC requirements
        def resp_rest(res: Response) -> Optional[TenantOutputType]:
            if res.status_code == 404:
                return None
            data = res.json()
            return TenantOutput(
                name=data["name"],
                activity_status=TenantActivityStatus(data["activityStatus"]),
            )

        return executor.execute(
            response_callback=resp_rest,
            method=self._connection.get,
            path=f"/schema/{self._name}/tenants/{tenant_name}",
            error_msg=f"Could not get tenant {tenant_name} for collection {self._name}",
            status_codes=_ExpectedStatusCodes(
                ok_in=[200, 404],
                error=f"Get tenant {tenant_name} for collection {self._name}",
            ),
        )

    def __update(
        self,
        tenants: Union[TenantUpdateInputType, Sequence[TenantUpdateInputType]],
    ) -> executor.Result[None]:
        path = "/schema/" + self._name + "/tenants"
        if isinstance(self._connection, ConnectionAsync):

            async def _execute() -> None:
                await asyncio.gather(
                    *[
                        executor.aresult(
                            self._connection.put(
                                path=path,
                                weaviate_object=mapped_tenants,
                                error_msg=f"Collection tenants may not have been updated properly for {self._name}",
                                status_codes=_ExpectedStatusCodes(
                                    ok_in=200,
                                    error=f"Update collection tenants for {self._name}",
                                ),
                            )
                        )
                        for mapped_tenants in self.__map_update_tenants(tenants)
                    ]
                )

            return _execute()
        for mapped_tenants in self.__map_update_tenants(tenants):
            self._connection.put(
                path=path,
                weaviate_object=mapped_tenants,
                error_msg=f"Collection tenants may not have been updated properly for {self._name}",
                status_codes=_ExpectedStatusCodes(
                    ok_in=200, error=f"Update collection tenants for {self._name}"
                ),
            )

    def update(
        self,
        tenants: Union[TenantUpdateInputType, Sequence[TenantUpdateInputType]],
    ) -> executor.Result[None]:
        """Update the specified tenants for this collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Args:
            tenants: A tenant name, `wvc.config.tenants.Tenant` object, or a list of tenants names
                and/or `wvc.config.tenants.Tenant` objects to update for the given collection.

        Raises:
            weaviate.exceptions.WeaviateConnectionError: If the network connection to Weaviate fails.
            weaviate.exceptions.UnexpectedStatusCodeError: If Weaviate reports a non-OK status.
            weaviate.exceptions.WeaviateInvalidInputError: If `tenants` is not a list of `wvc.Tenant` objects.
        """
        if self._validate_arguments:
            _validate_input(
                _ValidateArgument(
                    expected=[
                        Tenant,
                        TenantUpdate,
                        Sequence[Union[Tenant, TenantUpdate]],
                    ],
                    name="tenants",
                    value=tenants,
                )
            )
        return self.__update(tenants=tenants)

    def exists(self, tenant: TenantInputType) -> executor.Result[bool]:
        """Check if a tenant exists for this collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Args:
            tenant: Tenant name or `wvc.config.tenants.Tenant` object to check for existence.

        Returns:
            `True` if the tenant exists, `False` otherwise.

        Raises:
            weaviate.exceptions.WeaviateConnectionError: If the network connection to Weaviate fails.
            weaviate.exceptions.UnexpectedStatusCodeError: If Weaviate reports a non-OK status.
        """
        self._connection._weaviate_version.check_is_at_least_1_25_0("The 'exists' method")
        if self._validate_arguments:
            _validate_input(
                _ValidateArgument(
                    expected=[str, Tenant],
                    name="tenant",
                    value=tenant,
                )
            )

        def resp(res: Response) -> bool:
            return res.status_code == 200

        tenant_name = tenant.name if isinstance(tenant, Tenant) else tenant
        path = "/schema/" + self._name + "/tenants/" + tenant_name
        return executor.execute(
            response_callback=resp,
            method=self._connection.head,
            path=path,
            error_msg=f"Could not check if tenant exists for {self._name}",
            status_codes=_ExpectedStatusCodes(
                ok_in=[200, 404], error=f"Check if tenant exists for {self._name}"
            ),  # allow 404 to perform bool check on response code
        )

    def __update_tenant_activity_status(
        self,
        tenant: Union[TenantInputType, Sequence[TenantInputType]],
        activity_status: TenantUpdateActivityStatus,
    ) -> executor.Result[None]:
        if self._validate_arguments:
            _validate_input(
                _ValidateArgument(
                    expected=[
                        str,
                        Tenant,
                        Sequence[Union[str, Tenant]],
                    ],
                    name="tenant",
                    value=tenant,
                )
            )
        if isinstance(tenant, str) or isinstance(tenant, Tenant):
            tenants = [
                TenantUpdate(
                    name=tenant.name if isinstance(tenant, Tenant) else tenant,
                    activity_status=activity_status,
                )
            ]
        else:
            tenants = [
                TenantUpdate(
                    name=t.name if isinstance(t, Tenant) else t,
                    activity_status=activity_status,
                )
                for t in tenant
            ]
        return self.__update(tenants=tenants)

    def activate(
        self, tenant: Union[TenantInputType, Sequence[TenantInputType]]
    ) -> executor.Result[None]:
        """Activate the specified tenants for this collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Args:
            tenant: A tenant name, `wvc.config.tenants.Tenant` object, or a list of tenants names
                and/or `wvc.config.tenants.Tenant` objects to activate for the given collection.

        Raises:
            weaviate.exceptions.WeaviateConnectionError: If the network connection to Weaviate fails.
            weaviate.exceptions.UnexpectedStatusCodeError: If Weaviate reports a non-OK status.
            weaviate.exceptions.WeaviateInvalidInputError: If `tenant` is not a list of `wvc.Tenant` objects.
        """
        self.__update_tenant_activity_status(
            tenant=tenant,
            activity_status=TenantUpdateActivityStatus.ACTIVE,
        )

    def deactivate(
        self, tenant: Union[TenantInputType, Sequence[TenantInputType]]
    ) -> executor.Result[None]:
        """Deactivate the specified tenants for this collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Args:
            tenant: A tenant name, `wvc.config.tenants.Tenant` object, or a list of tenants names
                and/or `wvc.config.tenants.Tenant` objects to deactivate for the given collection.

        Raises:
            weaviate.exceptions.WeaviateConnectionError: If the network connection to Weaviate fails.
            weaviate.exceptions.UnexpectedStatusCodeError: If Weaviate reports a non-OK status.
            weaviate.exceptions.WeaviateInvalidInputError: If `tenant` is not a list of `wvc.Tenant` objects.
        """
        self.__update_tenant_activity_status(
            tenant=tenant,
            activity_status=TenantUpdateActivityStatus.INACTIVE,
        )

    def offload(
        self, tenant: Union[TenantInputType, Sequence[TenantInputType]]
    ) -> executor.Result[None]:
        """Offload the specified tenants for this collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Args:
            tenant: A tenant name, `wvc.config.tenants.Tenant` object, or a list of tenants names
                and/or `wvc.config.tenants.Tenant` objects to offload for the given collection.

        Raises:
            weaviate.exceptions.WeaviateConnectionError: If the network connection to Weaviate fails.
            weaviate.exceptions.UnexpectedStatusCodeError: If Weaviate reports a non-OK status.
            weaviate.exceptions.WeaviateInvalidInputError: If `tenant` is not a list of `wvc.Tenant` objects.
        """
        self.__update_tenant_activity_status(
            tenant=tenant,
            activity_status=TenantUpdateActivityStatus.OFFLOADED,
        )



from .generate import _FetchObjectsGenerate, _FetchObjectsGenerateAsync
from .query import _FetchObjectsQuery, _FetchObjectsQueryAsync

__all__ = [
    "_FetchObjectsGenerate",
    "_FetchObjectsGenerateAsync",
    "_FetchObjectsQuery",
    "_FetchObjectsQueryAsync",
]



from weaviate.cluster.models import ShardingState, ShardReplicas
from weaviate.collections.classes.cluster import (
    Node,
    NodeMinimal,
    NodeVerbose,
    Shard,
    Shards,
    Stats,
)

__all__ = [
    "Node",
    "NodeMinimal",
    "NodeVerbose",
    "Shard",
    "Shards",
    "Stats",
    "ShardingState",
    "ShardReplicas",
]



"""Authentication class definitions."""

from dataclasses import dataclass
from typing import List, Optional, Union

from weaviate.warnings import _Warnings

SCOPES = Union[str, List[str]]


@dataclass
class _ClientCredentials:
    """Authenticate for the Client Credential flow using client secrets.

    Acquire the client secret from your identify provider and set the appropriate scope. The client includes hardcoded
    scopes for Azure, otherwise it needs to be supplied.
    Scopes can be given as:
      - List of strings: ["scope1", "scope2"]
      - space separated string: "scope1 scope2"
    """

    client_secret: str
    scope: Optional[SCOPES] = None

    def __post_init__(self) -> None:
        if self.scope is None:
            self.scope_list: List[str] = []
        elif isinstance(self.scope, str):
            self.scope_list = self.scope.split(" ")
        elif isinstance(self.scope, list):
            self.scope_list = self.scope


@dataclass
class _ClientPassword:
    """Using username and password for authentication with Resource Owner Password flow.

    For some providers the scope needs to contain "offline_access" (and "openid" which is automatically added) to return
    a refresh token. Without a refresh token the authentication will expire once the lifetime of the access token is up.
    Scopes can be given as:
      - List of strings: ["scope1", "scope2"]
      - space separated string: "scope1 scope2"
    """

    username: str
    password: str
    scope: Optional[SCOPES] = None

    def __post_init__(self) -> None:
        if self.scope is None:
            self.scope_list: List[str] = []
        elif isinstance(self.scope, str):
            self.scope_list = self.scope.split(" ")
        elif isinstance(self.scope, list):
            self.scope_list = self.scope


@dataclass
class _BearerToken:
    """Using a preexisting bearer/access token for authentication.

    The expiration time of access tokens is given in seconds.

    Only the access token is required. However, when no refresh token is
    given, the authentication will expire once the lifetime of the
    access token is up.
    """

    access_token: str
    expires_in: int = 60
    refresh_token: Optional[str] = None

    def __post_init__(self) -> None:
        if self.expires_in and self.expires_in < 0:
            _Warnings.auth_negative_expiration_time(self.expires_in)


@dataclass
class _APIKey:
    """Using the given API key to authenticate with weaviate."""

    api_key: str


class Auth:
    @staticmethod
    def api_key(api_key: str) -> _APIKey:
        return _APIKey(api_key)

    @staticmethod
    def client_credentials(
        client_secret: str, scope: Optional[SCOPES] = None
    ) -> _ClientCredentials:
        return _ClientCredentials(client_secret, scope)

    @staticmethod
    def client_password(
        username: str, password: str, scope: Optional[SCOPES] = None
    ) -> _ClientPassword:
        return _ClientPassword(username=username, password=password, scope=scope)

    @staticmethod
    def bearer_token(
        access_token: str, expires_in: int = 60, refresh_token: Optional[str] = None
    ) -> _BearerToken:
        return _BearerToken(
            access_token=access_token,
            expires_in=expires_in,
            refresh_token=refresh_token,
        )


OidcAuth = Union[_BearerToken, _ClientPassword, _ClientCredentials]
AuthCredentials = Union[OidcAuth, _APIKey]

# required to ease v3 -> v4 transition
AuthApiKey = _APIKey
"""
.. deprecated:: 4.0.0
    Use :meth:`~weaviate.auth.Auth.api_key` instead.
"""
AuthBearerToken = _BearerToken
"""
.. deprecated:: 4.0.0
    Use :meth:`~weaviate.auth.Auth.bearer_token` instead.
"""
AuthClientCredentials = _ClientCredentials
"""
.. deprecated:: 4.0.0
    Use :meth:`~weaviate.auth.Auth.client_credentials` instead.
"""
AuthClientPassword = _ClientPassword
"""
.. deprecated:: 4.0.0
    Use :meth:`~weaviate.auth.Auth.client_password` instead.
"""



import datetime
import random
import uuid as uuid_lib
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import weaviate
import weaviate.classes as wvc

schema = {
    "classes": [
        {
            "class": "Author",
            "properties": [{"dataType": ["string"], "name": "name"}],
            "vectorizer": "none",
        },
        {
            "class": "Paragraph",
            "properties": [
                {"dataType": ["text"], "name": "contents"},
                {"dataType": ["Paragraph"], "name": "hasParagraphs"},
                {"dataType": ["Author"], "name": "author"},
            ],
            "vectorizer": "none",
        },
        {
            "class": "Article",
            "properties": [
                {"dataType": ["string"], "name": "title"},
                {"dataType": ["Paragraph"], "name": "hasParagraphs"},
                {"dataType": ["date"], "name": "datePublished"},
                {"dataType": ["Author"], "name": "author"},
                {"dataType": ["string"], "name": "somestring"},
                {"dataType": ["int"], "name": "counter"},
            ],
            "vectorizer": "none",
        },
    ]
}


@dataclass(frozen=True)
class Reference:
    to_class: str
    to_uuid: uuid_lib.UUID


@dataclass
class DataObject:
    properties: Dict[str, Any]
    class_name: str
    uuid: uuid_lib.UUID


@dataclass
class Author:
    name: str
    uuid: uuid_lib.UUID = field(init=False)
    class_name: str = field(init=False)

    def to_data_object(self) -> DataObject:
        return DataObject({"name": self.name}, self.class_name, self.uuid)

    def __post_init__(self) -> None:
        self.uuid = uuid_lib.uuid4()
        self.class_name = "Author"


@dataclass
class Paragraph:
    contents: str
    author: Reference
    hasParagraphs: Optional[Reference]
    uuid: uuid_lib.UUID = field(init=False)
    class_name: str = field(init=False)

    def to_data_object(self) -> DataObject:
        return DataObject({"contents": self.contents}, self.class_name, self.uuid)

    def __post_init__(self) -> None:
        self.uuid = uuid_lib.uuid4()
        self.class_name = "Paragraph"


@dataclass
class Article:
    title: str
    datePublished: str
    somestring: str
    counter: int
    author: Reference
    hasParagraphs: Reference
    uuid: uuid_lib.UUID = field(init=False)
    class_name: str = field(init=False)

    def to_data_object(self) -> DataObject:
        return DataObject(
            {"title": self.title, "datePublished": self.datePublished}, self.class_name, self.uuid
        )

    def __post_init__(self) -> None:
        self.uuid = uuid_lib.uuid4()
        self.class_name = "Article"


def test_stress() -> None:
    random.seed(0)
    client = weaviate.connect_to_local()
    for col in schema["classes"]:
        client.collections.delete(str(col["class"]))

    authors = client.collections.create_from_dict(schema["classes"][0])
    paragraphs = client.collections.create_from_dict(schema["classes"][1])
    articles = client.collections.create_from_dict(schema["classes"][2])

    author_data = create_authors(100)
    paragraph_data = create_paragraphs(num_paragraphs=200, authors=author_data)
    article_data = create_articles(3000, author_data, paragraph_data)

    add_authors(client, author_data)
    add_paragraphs(client, paragraph_data)
    add_articles(client, article_data)

    assert len(authors) == len(author_data)
    assert len(paragraphs) == len(paragraph_data)
    assert len(articles) == len(article_data)

    # verify references
    for article in article_data:
        article_weav = articles.query.fetch_object_by_id(
            article.uuid,
            return_references=[
                wvc.query.QueryReference(link_on="hasParagraphs"),
                wvc.query.QueryReference(link_on="author"),
            ],
        )
        assert article_weav.uuid == article.uuid
        assert (
            article_weav.references["hasParagraphs"].objects[0].uuid
            == article.hasParagraphs.to_uuid
        )
        assert article_weav.references["author"].objects[0].uuid == article.author.to_uuid

    for i, paragraph in enumerate(paragraph_data):
        para_weav = paragraphs.query.fetch_object_by_id(
            paragraph.uuid,
            return_references=[
                wvc.query.QueryReference(link_on="hasParagraphs"),
                wvc.query.QueryReference(link_on="author"),
            ],
        )
        assert para_weav.uuid == paragraph.uuid
        assert "author" in para_weav.references, i
        assert para_weav.references["author"].objects[0].uuid == paragraph.author.to_uuid, i
        if paragraph.hasParagraphs is not None:
            assert (
                para_weav.references["hasParagraphs"].objects[0].uuid
                == paragraph.hasParagraphs.to_uuid
            )
        else:
            assert "hasParagraphs" not in para_weav.references

    for col in schema["classes"]:
        client.collections.delete(str(col["class"]))


def add_authors(client: weaviate.WeaviateClient, authors: List[Author]) -> None:
    with client.batch.dynamic() as batch:
        for author in authors:
            data_object = author.to_data_object()
            batch.add_object(
                collection=data_object.class_name,
                properties=data_object.properties,
                uuid=data_object.uuid,
            )
    assert len(client.batch.failed_objects) == 0


def add_paragraphs(client: weaviate.WeaviateClient, paragraphs: List[Paragraph]) -> None:
    with client.batch.dynamic() as batch:
        for paragraph in paragraphs:
            data_object = paragraph.to_data_object()
            batch.add_object(
                collection=data_object.class_name,
                properties=data_object.properties,
                uuid=data_object.uuid,
            )
            batch.add_reference(
                from_uuid=str(paragraph.uuid),
                from_property="author",
                to=str(paragraph.author.to_uuid),
                from_collection="Paragraph",
            )
            if paragraph.hasParagraphs is not None:
                batch.add_reference(
                    from_uuid=str(paragraph.uuid),
                    from_property="hasParagraphs",
                    to=str(paragraph.hasParagraphs.to_uuid),
                    from_collection="Paragraph",
                )
    assert len(client.batch.failed_references) == 0
    assert len(client.batch.failed_objects) == 0


def add_articles(client: weaviate.WeaviateClient, articles: List[Article]) -> None:
    with client.batch.dynamic() as batch:
        for article in articles:
            data_object = article.to_data_object()
            batch.add_object(
                collection=data_object.class_name,
                properties=data_object.properties,
                uuid=data_object.uuid,
            )
            batch.add_reference(
                str(article.uuid),
                from_property="author",
                to=str(article.author.to_uuid),
                from_collection="Article",
            )
            batch.add_reference(
                str(article.uuid),
                from_property="hasParagraphs",
                to=str(article.hasParagraphs.to_uuid),
                from_collection="Article",
            )
    assert len(client.batch.failed_references) == 0
    assert len(client.batch.failed_objects) == 0


def create_authors(num_authors: int) -> List[Author]:
    authors: List[Author] = [Author(f"{i}") for i in range(num_authors)]
    return authors


def create_paragraphs(num_paragraphs: int, authors: List[Author]) -> List[Paragraph]:
    paragraphs: List[Paragraph] = []
    for i in range(num_paragraphs):
        content: str = f"{i} {i} {i} {i}"

        paragraph_to_reference: Optional[Paragraph] = None
        if len(paragraphs) > 0 and i % 2 == 0:
            paragraph_to_reference = paragraphs[i % len(paragraphs)]
        author_to_reference: Author = authors[0]
        paragraphs.append(
            Paragraph(
                content,
                Reference("Author", author_to_reference.uuid),
                (
                    Reference("Paragraph", paragraph_to_reference.uuid)
                    if paragraph_to_reference is not None
                    else None
                ),
            )
        )
    return paragraphs


def create_articles(
    num_articles: int, authors: List[Author], paragraphs: List[Paragraph]
) -> List[Article]:
    articles: List[Article] = []
    base_date: datetime.date = datetime.datetime(2023, 12, 9, 7, 1, 34)
    for i in range(num_articles):
        title: str = f"{i} {i} {i}"
        paragraph_to_reference: Paragraph = paragraphs[i % len(paragraphs)]
        author_to_reference: Author = authors[i % len(authors)]
        date_published: str = (base_date + datetime.timedelta(hours=i)).isoformat() + "Z"
        articles.append(
            Article(
                title,
                date_published,
                str(i),
                i,
                Reference("Author", author_to_reference.uuid),
                Reference("Paragraph", paragraph_to_reference.uuid),
            )
        )

    return articles



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v1/base_search.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from weaviate.proto.v1.v5261.v1 import base_pb2 as v1_dot_base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14v1/base_search.proto\x12\x0bweaviate.v1\x1a\rv1/base.proto\"2\n\x10WeightsForTarget\x12\x0e\n\x06target\x18\x01 \x01(\t\x12\x0e\n\x06weight\x18\x02 \x01(\x02\"\x98\x01\n\x07Targets\x12\x16\n\x0etarget_vectors\x18\x01 \x03(\t\x12\x33\n\x0b\x63ombination\x18\x02 \x01(\x0e\x32\x1e.weaviate.v1.CombinationMethod\x12:\n\x13weights_for_targets\x18\x04 \x03(\x0b\x32\x1d.weaviate.v1.WeightsForTargetJ\x04\x08\x03\x10\x04\"`\n\x0fVectorForTarget\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x18\n\x0cvector_bytes\x18\x02 \x01(\x0c\x42\x02\x18\x01\x12%\n\x07vectors\x18\x03 \x03(\x0b\x32\x14.weaviate.v1.Vectors\"\xe1\x01\n\x15SearchOperatorOptions\x12=\n\x08operator\x18\x01 \x01(\x0e\x32+.weaviate.v1.SearchOperatorOptions.Operator\x12$\n\x17minimum_or_tokens_match\x18\x02 \x01(\x05H\x00\x88\x01\x01\"G\n\x08Operator\x12\x18\n\x14OPERATOR_UNSPECIFIED\x10\x00\x12\x0f\n\x0bOPERATOR_OR\x10\x01\x12\x10\n\x0cOPERATOR_AND\x10\x02\x42\x1a\n\x18_minimum_or_tokens_match\"\xd0\x04\n\x06Hybrid\x12\r\n\x05query\x18\x01 \x01(\t\x12\x12\n\nproperties\x18\x02 \x03(\t\x12\x12\n\x06vector\x18\x03 \x03(\x02\x42\x02\x18\x01\x12\r\n\x05\x61lpha\x18\x04 \x01(\x02\x12\x33\n\x0b\x66usion_type\x18\x05 \x01(\x0e\x32\x1e.weaviate.v1.Hybrid.FusionType\x12\x18\n\x0cvector_bytes\x18\x06 \x01(\x0c\x42\x02\x18\x01\x12\x1a\n\x0etarget_vectors\x18\x07 \x03(\tB\x02\x18\x01\x12.\n\tnear_text\x18\x08 \x01(\x0b\x32\x1b.weaviate.v1.NearTextSearch\x12,\n\x0bnear_vector\x18\t \x01(\x0b\x32\x17.weaviate.v1.NearVector\x12%\n\x07targets\x18\n \x01(\x0b\x32\x14.weaviate.v1.Targets\x12\x45\n\x14\x62m25_search_operator\x18\x0b \x01(\x0b\x32\".weaviate.v1.SearchOperatorOptionsH\x01\x88\x01\x01\x12\x19\n\x0fvector_distance\x18\x14 \x01(\x02H\x00\x12%\n\x07vectors\x18\x15 \x03(\x0b\x32\x14.weaviate.v1.Vectors\"a\n\nFusionType\x12\x1b\n\x17\x46USION_TYPE_UNSPECIFIED\x10\x00\x12\x16\n\x12\x46USION_TYPE_RANKED\x10\x01\x12\x1e\n\x1a\x46USION_TYPE_RELATIVE_SCORE\x10\x02\x42\x0b\n\tthresholdB\x17\n\x15_bm25_search_operator\"\xad\x03\n\nNearVector\x12\x12\n\x06vector\x18\x01 \x03(\x02\x42\x02\x18\x01\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x12\x18\n\x0cvector_bytes\x18\x04 \x01(\x0c\x42\x02\x18\x01\x12\x1a\n\x0etarget_vectors\x18\x05 \x03(\tB\x02\x18\x01\x12%\n\x07targets\x18\x06 \x01(\x0b\x32\x14.weaviate.v1.Targets\x12K\n\x11vector_per_target\x18\x07 \x03(\x0b\x32,.weaviate.v1.NearVector.VectorPerTargetEntryB\x02\x18\x01\x12\x38\n\x12vector_for_targets\x18\x08 \x03(\x0b\x32\x1c.weaviate.v1.VectorForTarget\x12%\n\x07vectors\x18\t \x03(\x0b\x32\x14.weaviate.v1.Vectors\x1a\x36\n\x14VectorPerTargetEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x42\x0c\n\n_certaintyB\x0b\n\t_distance\"\xa5\x01\n\nNearObject\x12\n\n\x02id\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x12\x1a\n\x0etarget_vectors\x18\x04 \x03(\tB\x02\x18\x01\x12%\n\x07targets\x18\x05 \x01(\x0b\x32\x14.weaviate.v1.TargetsB\x0c\n\n_certaintyB\x0b\n\t_distance\"\xf0\x02\n\x0eNearTextSearch\x12\r\n\x05query\x18\x01 \x03(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x12\x36\n\x07move_to\x18\x04 \x01(\x0b\x32 .weaviate.v1.NearTextSearch.MoveH\x02\x88\x01\x01\x12\x38\n\tmove_away\x18\x05 \x01(\x0b\x32 .weaviate.v1.NearTextSearch.MoveH\x03\x88\x01\x01\x12\x1a\n\x0etarget_vectors\x18\x06 \x03(\tB\x02\x18\x01\x12%\n\x07targets\x18\x07 \x01(\x0b\x32\x14.weaviate.v1.Targets\x1a\x36\n\x04Move\x12\r\n\x05\x66orce\x18\x01 \x01(\x02\x12\x10\n\x08\x63oncepts\x18\x02 \x03(\t\x12\r\n\x05uuids\x18\x03 \x03(\tB\x0c\n\n_certaintyB\x0b\n\t_distanceB\n\n\x08_move_toB\x0c\n\n_move_away\"\xad\x01\n\x0fNearImageSearch\x12\r\n\x05image\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x12\x1a\n\x0etarget_vectors\x18\x04 \x03(\tB\x02\x18\x01\x12%\n\x07targets\x18\x05 \x01(\x0b\x32\x14.weaviate.v1.TargetsB\x0c\n\n_certaintyB\x0b\n\t_distance\"\xad\x01\n\x0fNearAudioSearch\x12\r\n\x05\x61udio\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x12\x1a\n\x0etarget_vectors\x18\x04 \x03(\tB\x02\x18\x01\x12%\n\x07targets\x18\x05 \x01(\x0b\x32\x14.weaviate.v1.TargetsB\x0c\n\n_certaintyB\x0b\n\t_distance\"\xad\x01\n\x0fNearVideoSearch\x12\r\n\x05video\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x12\x1a\n\x0etarget_vectors\x18\x04 \x03(\tB\x02\x18\x01\x12%\n\x07targets\x18\x05 \x01(\x0b\x32\x14.weaviate.v1.TargetsB\x0c\n\n_certaintyB\x0b\n\t_distance\"\xad\x01\n\x0fNearDepthSearch\x12\r\n\x05\x64\x65pth\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x12\x1a\n\x0etarget_vectors\x18\x04 \x03(\tB\x02\x18\x01\x12%\n\x07targets\x18\x05 \x01(\x0b\x32\x14.weaviate.v1.TargetsB\x0c\n\n_certaintyB\x0b\n\t_distance\"\xb1\x01\n\x11NearThermalSearch\x12\x0f\n\x07thermal\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x12\x1a\n\x0etarget_vectors\x18\x04 \x03(\tB\x02\x18\x01\x12%\n\x07targets\x18\x05 \x01(\x0b\x32\x14.weaviate.v1.TargetsB\x0c\n\n_certaintyB\x0b\n\t_distance\"\xa9\x01\n\rNearIMUSearch\x12\x0b\n\x03imu\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x12\x1a\n\x0etarget_vectors\x18\x04 \x03(\tB\x02\x18\x01\x12%\n\x07targets\x18\x05 \x01(\x0b\x32\x14.weaviate.v1.TargetsB\x0c\n\n_certaintyB\x0b\n\t_distance\"\x7f\n\x04\x42M25\x12\r\n\x05query\x18\x01 \x01(\t\x12\x12\n\nproperties\x18\x02 \x03(\t\x12@\n\x0fsearch_operator\x18\x03 \x01(\x0b\x32\".weaviate.v1.SearchOperatorOptionsH\x00\x88\x01\x01\x42\x12\n\x10_search_operator*\xee\x01\n\x11\x43ombinationMethod\x12\"\n\x1e\x43OMBINATION_METHOD_UNSPECIFIED\x10\x00\x12\x1f\n\x1b\x43OMBINATION_METHOD_TYPE_SUM\x10\x01\x12\x1f\n\x1b\x43OMBINATION_METHOD_TYPE_MIN\x10\x02\x12#\n\x1f\x43OMBINATION_METHOD_TYPE_AVERAGE\x10\x03\x12*\n&COMBINATION_METHOD_TYPE_RELATIVE_SCORE\x10\x04\x12\"\n\x1e\x43OMBINATION_METHOD_TYPE_MANUAL\x10\x05\x42t\n#io.weaviate.client.grpc.protocol.v1B\x17WeaviateProtoBaseSearchZ4github.com/weaviate/weaviate/grpc/generated;protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'v1.base_search_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n#io.weaviate.client.grpc.protocol.v1B\027WeaviateProtoBaseSearchZ4github.com/weaviate/weaviate/grpc/generated;protocol'
  _globals['_VECTORFORTARGET'].fields_by_name['vector_bytes']._loaded_options = None
  _globals['_VECTORFORTARGET'].fields_by_name['vector_bytes']._serialized_options = b'\030\001'
  _globals['_HYBRID'].fields_by_name['vector']._loaded_options = None
  _globals['_HYBRID'].fields_by_name['vector']._serialized_options = b'\030\001'
  _globals['_HYBRID'].fields_by_name['vector_bytes']._loaded_options = None
  _globals['_HYBRID'].fields_by_name['vector_bytes']._serialized_options = b'\030\001'
  _globals['_HYBRID'].fields_by_name['target_vectors']._loaded_options = None
  _globals['_HYBRID'].fields_by_name['target_vectors']._serialized_options = b'\030\001'
  _globals['_NEARVECTOR_VECTORPERTARGETENTRY']._loaded_options = None
  _globals['_NEARVECTOR_VECTORPERTARGETENTRY']._serialized_options = b'8\001'
  _globals['_NEARVECTOR'].fields_by_name['vector']._loaded_options = None
  _globals['_NEARVECTOR'].fields_by_name['vector']._serialized_options = b'\030\001'
  _globals['_NEARVECTOR'].fields_by_name['vector_bytes']._loaded_options = None
  _globals['_NEARVECTOR'].fields_by_name['vector_bytes']._serialized_options = b'\030\001'
  _globals['_NEARVECTOR'].fields_by_name['target_vectors']._loaded_options = None
  _globals['_NEARVECTOR'].fields_by_name['target_vectors']._serialized_options = b'\030\001'
  _globals['_NEARVECTOR'].fields_by_name['vector_per_target']._loaded_options = None
  _globals['_NEARVECTOR'].fields_by_name['vector_per_target']._serialized_options = b'\030\001'
  _globals['_NEAROBJECT'].fields_by_name['target_vectors']._loaded_options = None
  _globals['_NEAROBJECT'].fields_by_name['target_vectors']._serialized_options = b'\030\001'
  _globals['_NEARTEXTSEARCH'].fields_by_name['target_vectors']._loaded_options = None
  _globals['_NEARTEXTSEARCH'].fields_by_name['target_vectors']._serialized_options = b'\030\001'
  _globals['_NEARIMAGESEARCH'].fields_by_name['target_vectors']._loaded_options = None
  _globals['_NEARIMAGESEARCH'].fields_by_name['target_vectors']._serialized_options = b'\030\001'
  _globals['_NEARAUDIOSEARCH'].fields_by_name['target_vectors']._loaded_options = None
  _globals['_NEARAUDIOSEARCH'].fields_by_name['target_vectors']._serialized_options = b'\030\001'
  _globals['_NEARVIDEOSEARCH'].fields_by_name['target_vectors']._loaded_options = None
  _globals['_NEARVIDEOSEARCH'].fields_by_name['target_vectors']._serialized_options = b'\030\001'
  _globals['_NEARDEPTHSEARCH'].fields_by_name['target_vectors']._loaded_options = None
  _globals['_NEARDEPTHSEARCH'].fields_by_name['target_vectors']._serialized_options = b'\030\001'
  _globals['_NEARTHERMALSEARCH'].fields_by_name['target_vectors']._loaded_options = None
  _globals['_NEARTHERMALSEARCH'].fields_by_name['target_vectors']._serialized_options = b'\030\001'
  _globals['_NEARIMUSEARCH'].fields_by_name['target_vectors']._loaded_options = None
  _globals['_NEARIMUSEARCH'].fields_by_name['target_vectors']._serialized_options = b'\030\001'
  _globals['_COMBINATIONMETHOD']._serialized_start=3337
  _globals['_COMBINATIONMETHOD']._serialized_end=3575
  _globals['_WEIGHTSFORTARGET']._serialized_start=52
  _globals['_WEIGHTSFORTARGET']._serialized_end=102
  _globals['_TARGETS']._serialized_start=105
  _globals['_TARGETS']._serialized_end=257
  _globals['_VECTORFORTARGET']._serialized_start=259
  _globals['_VECTORFORTARGET']._serialized_end=355
  _globals['_SEARCHOPERATOROPTIONS']._serialized_start=358
  _globals['_SEARCHOPERATOROPTIONS']._serialized_end=583
  _globals['_SEARCHOPERATOROPTIONS_OPERATOR']._serialized_start=484
  _globals['_SEARCHOPERATOROPTIONS_OPERATOR']._serialized_end=555
  _globals['_HYBRID']._serialized_start=586
  _globals['_HYBRID']._serialized_end=1178
  _globals['_HYBRID_FUSIONTYPE']._serialized_start=1043
  _globals['_HYBRID_FUSIONTYPE']._serialized_end=1140
  _globals['_NEARVECTOR']._serialized_start=1181
  _globals['_NEARVECTOR']._serialized_end=1610
  _globals['_NEARVECTOR_VECTORPERTARGETENTRY']._serialized_start=1529
  _globals['_NEARVECTOR_VECTORPERTARGETENTRY']._serialized_end=1583
  _globals['_NEAROBJECT']._serialized_start=1613
  _globals['_NEAROBJECT']._serialized_end=1778
  _globals['_NEARTEXTSEARCH']._serialized_start=1781
  _globals['_NEARTEXTSEARCH']._serialized_end=2149
  _globals['_NEARTEXTSEARCH_MOVE']._serialized_start=2042
  _globals['_NEARTEXTSEARCH_MOVE']._serialized_end=2096
  _globals['_NEARIMAGESEARCH']._serialized_start=2152
  _globals['_NEARIMAGESEARCH']._serialized_end=2325
  _globals['_NEARAUDIOSEARCH']._serialized_start=2328
  _globals['_NEARAUDIOSEARCH']._serialized_end=2501
  _globals['_NEARVIDEOSEARCH']._serialized_start=2504
  _globals['_NEARVIDEOSEARCH']._serialized_end=2677
  _globals['_NEARDEPTHSEARCH']._serialized_start=2680
  _globals['_NEARDEPTHSEARCH']._serialized_end=2853
  _globals['_NEARTHERMALSEARCH']._serialized_start=2856
  _globals['_NEARTHERMALSEARCH']._serialized_end=3033
  _globals['_NEARIMUSEARCH']._serialized_start=3036
  _globals['_NEARIMUSEARCH']._serialized_end=3205
  _globals['_BM25']._serialized_start=3207
  _globals['_BM25']._serialized_end=3334
# @@protoc_insertion_point(module_scope)



from typing import Generic

from weaviate.collections.classes.types import References, TProperties
from weaviate.collections.queries.bm25 import _BM25Generate, _BM25GenerateAsync
from weaviate.collections.queries.fetch_objects import (
    _FetchObjectsGenerate,
    _FetchObjectsGenerateAsync,
)
from weaviate.collections.queries.fetch_objects_by_ids import (
    _FetchObjectsByIDsGenerate,
    _FetchObjectsByIDsGenerateAsync,
)
from weaviate.collections.queries.hybrid import _HybridGenerate, _HybridGenerateAsync
from weaviate.collections.queries.near_image import (
    _NearImageGenerate,
    _NearImageGenerateAsync,
)
from weaviate.collections.queries.near_media import (
    _NearMediaGenerate,
    _NearMediaGenerateAsync,
)
from weaviate.collections.queries.near_object import (
    _NearObjectGenerate,
    _NearObjectGenerateAsync,
)
from weaviate.collections.queries.near_text import (
    _NearTextGenerate,
    _NearTextGenerateAsync,
)
from weaviate.collections.queries.near_vector import (
    _NearVectorGenerate,
    _NearVectorGenerateAsync,
)


class _GenerateCollectionAsync(
    Generic[TProperties, References],
    _BM25GenerateAsync[TProperties, References],
    _FetchObjectsGenerateAsync[TProperties, References],
    _FetchObjectsByIDsGenerateAsync[TProperties, References],
    _HybridGenerateAsync[TProperties, References],
    _NearImageGenerateAsync[TProperties, References],
    _NearMediaGenerateAsync[TProperties, References],
    _NearObjectGenerateAsync[TProperties, References],
    _NearTextGenerateAsync[TProperties, References],
    _NearVectorGenerateAsync[TProperties, References],
):
    pass


class _GenerateCollection(
    Generic[TProperties, References],
    _BM25Generate[TProperties, References],
    _FetchObjectsGenerate[TProperties, References],
    _FetchObjectsByIDsGenerate[TProperties, References],
    _HybridGenerate[TProperties, References],
    _NearImageGenerate[TProperties, References],
    _NearMediaGenerate[TProperties, References],
    _NearObjectGenerate[TProperties, References],
    _NearTextGenerate[TProperties, References],
    _NearVectorGenerate[TProperties, References],
):
    pass



from typing import Any, Dict, List, Optional, Union, cast

from weaviate.collections.classes.config import (
    DataType,
    GenerativeSearches,
    PQEncoderDistribution,
    PQEncoderType,
    ReplicationDeletionStrategy,
    Rerankers,
    StopwordsPreset,
    Tokenization,
    VectorDistances,
    VectorFilterStrategy,
    VectorIndexType,
    Vectorizers,
    _BM25Config,
    _BQConfig,
    _CollectionConfig,
    _CollectionConfigSimple,
    _GenerativeConfig,
    _InvertedIndexConfig,
    _MultiTenancyConfig,
    _MultiVectorConfig,
    _MuveraConfig,
    _NamedVectorConfig,
    _NamedVectorizerConfig,
    _NestedProperty,
    _PQConfig,
    _PQEncoderConfig,
    _Property,
    _PropertyVectorizerConfig,
    _ReferenceProperty,
    _ReplicationConfig,
    _RerankerConfig,
    _RQConfig,
    _ShardingConfig,
    _SQConfig,
    _StopwordsConfig,
    _VectorIndexConfigDynamic,
    _VectorIndexConfigFlat,
    _VectorIndexConfigHNSW,
    _VectorizerConfig,
)


def _is_primitive(d_type: str) -> bool:
    return d_type[0][0].lower() == d_type[0][0]


def __get_rerank_config(schema: Dict[str, Any]) -> Optional[_RerankerConfig]:
    if (
        len(
            rerankers := [key for key in schema.get("moduleConfig", {}).keys() if "reranker" in key]
        )
        == 1
    ):
        try:
            reranker = Rerankers(rerankers[0])
        except ValueError:
            reranker = rerankers[0]
        return _RerankerConfig(
            model=schema["moduleConfig"][rerankers[0]],
            reranker=reranker,
        )
    else:
        return None


def __get_generative_config(schema: Dict[str, Any]) -> Optional[_GenerativeConfig]:
    if (
        len(
            generators := [
                key for key in schema.get("moduleConfig", {}).keys() if "generative" in key
            ]
        )
        == 1
    ):
        try:
            generative = GenerativeSearches(generators[0])
        except ValueError:
            generative = generators[0]

        return _GenerativeConfig(
            generative=generative,
            model=schema["moduleConfig"][generators[0]],
        )
    else:
        return None


def __get_vectorizer_config(schema: Dict[str, Any]) -> Optional[_VectorizerConfig]:
    if __is_vectorizer_present(schema) is not None and schema.get("vectorizer", "none") != "none":
        vec_config: Dict[str, Any] = schema["moduleConfig"].pop(schema["vectorizer"])
        try:
            vectorizer = Vectorizers(schema["vectorizer"])
        except ValueError:
            vectorizer = schema["vectorizer"]
        return _VectorizerConfig(
            vectorize_collection_name=vec_config.pop("vectorizeClassName", False),
            model=vec_config,
            vectorizer=vectorizer,
        )
    else:
        return None


def __is_vectorizer_present(schema: Dict[str, Any]) -> bool:
    # ignore single vectorizer config if named vectors are present
    if "vectorConfig" in schema:
        return False
    return True


def __get_vector_index_type(schema: Dict[str, Any]) -> Optional[VectorIndexType]:
    if "vectorIndexType" in schema:
        return VectorIndexType(schema["vectorIndexType"])
    else:
        return None


def __get_quantizer_config(
    config: Dict[str, Any],
) -> Optional[Union[_PQConfig, _BQConfig, _SQConfig, _RQConfig]]:
    quantizer: Optional[Union[_PQConfig, _BQConfig, _SQConfig, _RQConfig]] = None
    if "bq" in config and config["bq"]["enabled"]:
        # values are not present for bq+hnsw
        quantizer = _BQConfig(
            cache=config["bq"].get("cache"),
            rescore_limit=config["bq"].get("rescoreLimit"),
        )
    elif "sq" in config and config["sq"]["enabled"]:
        # values are not present for bq+hnsw
        quantizer = _SQConfig(
            cache=config["sq"].get("cache"),
            rescore_limit=config["sq"].get("rescoreLimit"),
            training_limit=config["sq"].get("trainingLimit"),
        )
    elif "pq" in config and config["pq"].get("enabled"):
        quantizer = _PQConfig(
            internal_bit_compression=config["pq"].get("bitCompression"),
            segments=config["pq"].get("segments"),
            centroids=config["pq"].get("centroids"),
            training_limit=config["pq"].get("trainingLimit"),
            encoder=_PQEncoderConfig(
                type_=PQEncoderType(config["pq"].get("encoder", {}).get("type")),
                distribution=PQEncoderDistribution(
                    config["pq"].get("encoder", {}).get("distribution")
                ),
            ),
        )
    elif "rq" in config and config["rq"].get("enabled"):
        quantizer = _RQConfig(
            bits=config["rq"].get("bits"),
            rescore_limit=config["rq"].get("rescoreLimit"),
        )
    return quantizer


def __get_multivector_encoding(config: Dict[str, Any]) -> Optional[_MuveraConfig]:
    return (
        None
        if config.get("muvera") is None
        or not config.get("muvera", {"enabled": False}).get("enabled")
        else _MuveraConfig(
            enabled=config["muvera"]["enabled"],
            ksim=config["muvera"]["ksim"],
            dprojections=config["muvera"]["dprojections"],
            repetitions=config["muvera"]["repetitions"],
        )
    )


def __get_multivector(config: Dict[str, Any]) -> Optional[_MultiVectorConfig]:
    return (
        None
        if config.get("multivector") is None
        or not config.get("multivector", {"enabled": False}).get("enabled")
        else _MultiVectorConfig(
            encoding=(
                None
                if config["multivector"].get("muvera") is None
                else __get_multivector_encoding(config["multivector"])
            ),
            aggregation=config["multivector"]["aggregation"],
        )
    )


def __get_hnsw_config(config: Dict[str, Any]) -> _VectorIndexConfigHNSW:
    quantizer = __get_quantizer_config(config)
    return _VectorIndexConfigHNSW(
        cleanup_interval_seconds=config["cleanupIntervalSeconds"],
        distance_metric=VectorDistances(config.get("distance")),
        dynamic_ef_min=config["dynamicEfMin"],
        dynamic_ef_max=config["dynamicEfMax"],
        dynamic_ef_factor=config["dynamicEfFactor"],
        ef=config["ef"],
        ef_construction=config["efConstruction"],
        filter_strategy=(
            VectorFilterStrategy(config["filterStrategy"])
            if "filterStrategy" in config
            else VectorFilterStrategy.SWEEPING
        ),
        flat_search_cutoff=config["flatSearchCutoff"],
        max_connections=config["maxConnections"],
        quantizer=quantizer,
        skip=config["skip"],
        vector_cache_max_objects=config["vectorCacheMaxObjects"],
        multi_vector=__get_multivector(config),
    )


def __get_flat_config(config: Dict[str, Any]) -> _VectorIndexConfigFlat:
    quantizer = __get_quantizer_config(config)
    return _VectorIndexConfigFlat(
        distance_metric=VectorDistances(config["distance"]),
        quantizer=quantizer,
        vector_cache_max_objects=config["vectorCacheMaxObjects"],
        multi_vector=__get_multivector(config),
    )


def __get_vector_index_config(
    schema: Dict[str, Any],
) -> Union[_VectorIndexConfigHNSW, _VectorIndexConfigFlat, _VectorIndexConfigDynamic, None]:
    if "vectorIndexConfig" not in schema:
        return None
    if schema["vectorIndexType"] == "hnsw":
        return __get_hnsw_config(schema["vectorIndexConfig"])
    elif schema["vectorIndexType"] == "flat":
        return __get_flat_config(schema["vectorIndexConfig"])
    elif schema["vectorIndexType"] == "dynamic":
        return _VectorIndexConfigDynamic(
            distance_metric=VectorDistances(schema["vectorIndexConfig"]["distance"]),
            threshold=schema["vectorIndexConfig"].get("threshold"),
            hnsw=__get_hnsw_config(schema["vectorIndexConfig"]["hnsw"]),
            flat=__get_flat_config(schema["vectorIndexConfig"]["flat"]),
        )
    else:
        return None


def __get_vector_config(
    schema: Dict[str, Any], simple: bool
) -> Optional[Dict[str, _NamedVectorConfig]]:
    if "vectorConfig" in schema:
        named_vectors: Dict[str, _NamedVectorConfig] = {}
        for name in schema["vectorConfig"]:
            named_vector = schema["vectorConfig"][name]

            vectorizer = named_vector["vectorizer"].keys()
            assert len(vectorizer) == 1

            vectorizer_str: str = str(list(vectorizer)[0])
            vec_config: Dict[str, Any] = named_vector["vectorizer"][vectorizer_str]
            props = vec_config.pop("properties", None)

            vector_index_config = __get_vector_index_config(named_vector)
            assert vector_index_config is not None
            try:
                vec: Union[str, Vectorizers] = Vectorizers(vectorizer_str)
            except ValueError:
                vec = vectorizer_str

            named_vectors[name] = _NamedVectorConfig(
                vectorizer=_NamedVectorizerConfig(
                    vectorizer=vec,
                    model=vec_config,
                    source_properties=props,
                ),
                vector_index_config=vector_index_config,
            )

        return named_vectors
    else:
        return None


def __get_vectorizer(schema: Dict[str, Any]) -> Optional[Union[str, Vectorizers]]:
    if "vectorConfig" in schema:
        return None

    vectorizer = str(schema["vectorizer"])
    try:
        return Vectorizers(vectorizer)
    except ValueError:
        return vectorizer


def _collection_config_simple_from_json(schema: Dict[str, Any]) -> _CollectionConfigSimple:
    return _CollectionConfigSimple(
        name=schema["class"],
        description=schema.get("description"),
        generative_config=__get_generative_config(schema),
        properties=(
            _properties_from_config(schema) if schema.get("properties") is not None else []
        ),
        references=(
            _references_from_config(schema) if schema.get("properties") is not None else []
        ),
        reranker_config=__get_rerank_config(schema),
        vectorizer_config=__get_vectorizer_config(schema),
        vectorizer=__get_vectorizer(schema),
        vector_config=__get_vector_config(schema, simple=True),
    )


def _collection_config_from_json(schema: Dict[str, Any]) -> _CollectionConfig:
    return _CollectionConfig(
        name=schema["class"],
        description=schema.get("description"),
        generative_config=__get_generative_config(schema),
        inverted_index_config=_InvertedIndexConfig(
            bm25=_BM25Config(
                b=schema["invertedIndexConfig"]["bm25"]["b"],
                k1=schema["invertedIndexConfig"]["bm25"]["k1"],
            ),
            cleanup_interval_seconds=schema["invertedIndexConfig"]["cleanupIntervalSeconds"],
            index_null_state=cast(dict, schema["invertedIndexConfig"]).get("indexNullState")
            is True,
            index_property_length=cast(dict, schema["invertedIndexConfig"]).get(
                "indexPropertyLength"
            )
            is True,
            index_timestamps=cast(dict, schema["invertedIndexConfig"]).get("indexTimestamps")
            is True,
            stopwords=_StopwordsConfig(
                preset=StopwordsPreset(schema["invertedIndexConfig"]["stopwords"]["preset"]),
                additions=schema["invertedIndexConfig"]["stopwords"]["additions"],
                removals=schema["invertedIndexConfig"]["stopwords"]["removals"],
            ),
        ),
        multi_tenancy_config=_MultiTenancyConfig(
            enabled=schema.get("multiTenancyConfig", {}).get("enabled", False),
            auto_tenant_creation=schema.get("multiTenancyConfig", {}).get(
                "autoTenantCreation", False
            ),
            auto_tenant_activation=schema.get("multiTenancyConfig", {}).get(
                "autoTenantActivation", False
            ),
        ),
        properties=(
            _properties_from_config(schema) if schema.get("properties") is not None else []
        ),
        references=(
            _references_from_config(schema) if schema.get("properties") is not None else []
        ),
        replication_config=_ReplicationConfig(
            factor=schema["replicationConfig"]["factor"],
            async_enabled=schema["replicationConfig"].get("asyncEnabled", False),
            deletion_strategy=(
                ReplicationDeletionStrategy(schema["replicationConfig"]["deletionStrategy"])
                if "deletionStrategy" in schema["replicationConfig"]
                else ReplicationDeletionStrategy.NO_AUTOMATED_RESOLUTION
            ),
        ),
        reranker_config=__get_rerank_config(schema),
        sharding_config=(
            None
            if schema.get("multiTenancyConfig", {}).get("enabled", False)
            else _ShardingConfig(
                virtual_per_physical=schema["shardingConfig"]["virtualPerPhysical"],
                desired_count=schema["shardingConfig"]["desiredCount"],
                actual_count=schema["shardingConfig"]["actualCount"],
                desired_virtual_count=schema["shardingConfig"]["desiredVirtualCount"],
                actual_virtual_count=schema["shardingConfig"]["actualVirtualCount"],
                key=schema["shardingConfig"]["key"],
                strategy=schema["shardingConfig"]["strategy"],
                function=schema["shardingConfig"]["function"],
            )
        ),
        vector_index_config=__get_vector_index_config(schema),
        vector_index_type=__get_vector_index_type(schema),
        vectorizer_config=__get_vectorizer_config(schema),
        vectorizer=__get_vectorizer(schema),
        vector_config=__get_vector_config(schema, simple=False),
    )


def _collection_configs_from_json(schema: Dict[str, Any]) -> Dict[str, _CollectionConfig]:
    configs = {
        schema["class"]: _collection_config_from_json(schema) for schema in schema["classes"]
    }
    return dict(sorted(configs.items()))


def _collection_configs_simple_from_json(
    schema: Dict[str, Any],
) -> Dict[str, _CollectionConfigSimple]:
    configs = {
        schema["class"]: _collection_config_simple_from_json(schema) for schema in schema["classes"]
    }
    return dict(sorted(configs.items()))


def _nested_properties_from_config(props: List[Dict[str, Any]]) -> List[_NestedProperty]:
    return [
        _NestedProperty(
            data_type=DataType(prop["dataType"][0]),
            description=prop.get("description"),
            index_filterable=prop["indexFilterable"],
            index_searchable=prop["indexSearchable"],
            name=prop["name"],
            nested_properties=(
                _nested_properties_from_config(prop["nestedProperties"])
                if prop.get("nestedProperties") is not None
                else None
            ),
            tokenization=(
                Tokenization(prop["tokenization"]) if prop.get("tokenization") is not None else None
            ),
        )
        for prop in props
    ]


def _properties_from_config(schema: Dict[str, Any]) -> List[_Property]:
    return [
        _Property(
            data_type=DataType(prop["dataType"][0]),
            description=prop.get("description"),
            index_filterable=prop["indexFilterable"],
            index_range_filters=prop.get("indexRangeFilters", False),
            index_searchable=prop["indexSearchable"],
            name=prop["name"],
            nested_properties=(
                _nested_properties_from_config(prop["nestedProperties"])
                if prop.get("nestedProperties") is not None
                else None
            ),
            tokenization=(
                Tokenization(prop["tokenization"]) if prop.get("tokenization") is not None else None
            ),
            vectorizer_config=(
                _PropertyVectorizerConfig(
                    skip=prop["moduleConfig"][schema["vectorizer"]].get("skip", False),
                    vectorize_property_name=prop["moduleConfig"][schema["vectorizer"]].get(
                        "vectorizePropertyName", False
                    ),
                )
                if schema.get("vectorizer", "none") != "none"
                and prop.get("moduleConfig", None) is not None
                else None
            ),
            vectorizer_configs=(
                {
                    k: _PropertyVectorizerConfig(
                        skip=v.get("skip", False),
                        vectorize_property_name=v.get("vectorizePropertyName", False),
                    )
                    for k, v in prop.get("moduleConfig", {}).items()
                }
                if "vectorConfig" in schema
                else None
            ),
            vectorizer=(schema.get("vectorizer", "none") if "vectorConfig" not in schema else None),
        )
        for prop in schema["properties"]
        if _is_primitive(prop["dataType"])
    ]


def _references_from_config(schema: Dict[str, Any]) -> List[_ReferenceProperty]:
    return [
        _ReferenceProperty(
            target_collections=prop["dataType"],
            description=prop.get("description"),
            name=prop["name"],
        )
        for prop in schema["properties"]
        if not _is_primitive(prop["dataType"])
    ]



from weaviate.cluster.models import (
    ReplicateOperation,
    ReplicateOperations,
    ReplicateOperationState,
    ReplicateOperationStatus,
    ReplicateOperationWithHistory,
    ReplicateOperationWithoutHistory,
)

__all__ = [
    "ReplicateOperation",
    "ReplicateOperations",
    "ReplicateOperationWithHistory",
    "ReplicateOperationWithoutHistory",
    "ReplicateOperationState",
    "ReplicateOperationStatus",
]



import pathlib
import uuid
from datetime import datetime, timezone
from typing import Union

import pytest
from _pytest.fixtures import SubRequest

from integration.conftest import CollectionFactory, CollectionFactoryGet
from weaviate.collections.classes.aggregate import (
    AggregateBoolean,
    AggregateDate,
    AggregateInteger,
    AggregateNumber,
    AggregateReturn,
    AggregateText,
    GroupByAggregate,
    Metrics,
)
from weaviate.collections.classes.config import (
    Configure,
    DataType,
    Property,
    ReferenceProperty,
)
from weaviate.collections.classes.filters import Filter, _Filters
from weaviate.collections.classes.grpc import Move
from weaviate.collections.classes.tenants import Tenant
from weaviate.exceptions import (
    WeaviateInvalidInputError,
    WeaviateQueryError,
    WeaviateUnsupportedFeatureError,
)
from weaviate.util import file_encoder_b64

import weaviate.classes as wvc


UUID1 = uuid.UUID("8ad0d33c-8db1-4437-87f3-72161ca2a51a")
UUID2 = uuid.UUID("577887c1-4c6b-5594-aa62-f0c17883d9cf")


@pytest.mark.parametrize("how_many", [1, 10000, 20000, 20001, 100000])
def test_collection_length(collection_factory: CollectionFactory, how_many: int) -> None:
    """Uses `.aggregate` behind-the-scenes."""
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
    )
    collection.data.insert_many([{} for _ in range(how_many)])
    assert len(collection) == how_many


@pytest.mark.parametrize("how_many", [1, 10000, 20000, 20001, 100000])
def test_collection_length_tenant(collection_factory: CollectionFactory, how_many: int) -> None:
    """Uses `.aggregate` behind-the-scenes."""
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )
    collection.tenants.create(
        tenants=[Tenant(name="tenant1"), Tenant(name="tenant2"), Tenant(name="tenant3")]
    )
    collection.with_tenant("tenant2").data.insert_many([{} for _ in range(how_many * 2)])

    assert len(collection.with_tenant("tenant2")) == how_many * 2
    assert len(collection.with_tenant("tenant3")) == 0


def test_empty_aggregation(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(properties=[Property(name="text", data_type=DataType.TEXT)])
    res = collection.aggregate.over_all()
    assert res.total_count == 0


def test_simple_aggregation(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(properties=[Property(name="text", data_type=DataType.TEXT)])
    collection.data.insert({"text": "some text"})
    res = collection.aggregate.over_all(return_metrics=[Metrics("text").text(count=True)])
    assert isinstance(res.properties["text"], AggregateText)
    assert res.properties["text"].count == 1


def test_aggregation_top_occurence_with_limit(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(properties=[Property(name="text", data_type=DataType.TEXT)])
    collection.data.insert({"text": "one"})
    collection.data.insert({"text": "one"})
    collection.data.insert({"text": "two"})

    # Test deprecated min_occurrences parameter triggers warning
    with pytest.warns(DeprecationWarning) as recwarn:
        res_old = collection.aggregate.over_all(
            return_metrics=[Metrics("text").text(min_occurrences=1)],
        )
        assert len(recwarn) == 1
        assert "Dep028" in str(recwarn[0].message)
        assert "min_occurrences" in str(recwarn[0].message)
        assert "limit" in str(recwarn[0].message)

    # Test new limit parameter
    res = collection.aggregate.over_all(
        return_metrics=[Metrics("text").text(limit=1)],
    )

    # Verify both approaches produce the same result
    assert res.properties["text"] == res_old.properties["text"]

    # Test acutal output
    assert isinstance(res.properties["text"], AggregateText)
    assert len(res.properties["text"].top_occurrences) == 1
    assert res.properties["text"].top_occurrences[0].count == 2


def test_aggregation_groupby_with_limit(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(properties=[Property(name="text", data_type=DataType.TEXT)])
    collection.data.insert({"text": "one"})
    collection.data.insert({"text": "two"})
    collection.data.insert({"text": "three"})
    res = collection.aggregate.over_all(
        return_metrics=[Metrics("text").text(count=True)],
        group_by=GroupByAggregate(prop="text", limit=2),
    )
    assert len(res.groups) == 2
    assert res.groups[0].properties["text"].count == 1
    assert res.groups[1].properties["text"].count == 1


def test_aggregation_groupby_no_results(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(properties=[Property(name="text", data_type=DataType.TEXT)])
    res = collection.aggregate.over_all(
        return_metrics=[Metrics("text").text(count=True)],
        group_by=GroupByAggregate(prop="text", limit=2),
    )
    assert len(res.groups) == 0


@pytest.mark.parametrize(
    "filter_",
    [
        Filter.by_property("text").equal("two"),
        Filter.by_property("int").equal(2),
        Filter.by_property("float").equal(2.0),
        Filter.by_property("bool").equal(False),
        Filter.by_property("date").equal(datetime(2021, 1, 2, 0, 0, 0, 0, tzinfo=timezone.utc)),
        Filter.by_property("text").equal("two") | Filter.by_property("int").equal(2),
        Filter.by_property("uuid").equal(UUID2),
        Filter.by_property("texts").contains_any(["two"]),
        Filter.by_property("ints").contains_any([2]),
        Filter.by_property("floats").contains_any([2.0]),
        Filter.by_property("bools").contains_any([False]),
        Filter.by_property("dates").contains_any(
            [datetime(2021, 1, 2, 0, 0, 0, 0, tzinfo=timezone.utc)]
        ),
        Filter.by_property("uuids").contains_any([UUID2]),
    ],
)
def test_over_all_with_filters(collection_factory: CollectionFactory, filter_: _Filters) -> None:
    collection = collection_factory(
        properties=[
            Property(name="text", data_type=DataType.TEXT),
            Property(name="int", data_type=DataType.INT),
            Property(name="float", data_type=DataType.NUMBER),
            Property(name="bool", data_type=DataType.BOOL),
            Property(name="date", data_type=DataType.DATE),
            Property(name="uuid", data_type=DataType.UUID),
            Property(name="texts", data_type=DataType.TEXT_ARRAY),
            Property(name="ints", data_type=DataType.INT_ARRAY),
            Property(name="floats", data_type=DataType.NUMBER_ARRAY),
            Property(name="bools", data_type=DataType.BOOL_ARRAY),
            Property(name="dates", data_type=DataType.DATE_ARRAY),
            Property(name="uuids", data_type=DataType.UUID_ARRAY),
        ]
    )
    collection.data.insert(
        {
            "text": "one",
            "int": 1,
            "float": 1.0,
            "bool": True,
            "date": "2021-01-01T00:00:00Z",
            "uuid": UUID1,
            "texts": ["one"],
            "ints": [1],
            "floats": [1.0],
            "bools": [True],
            "dates": ["2021-01-01T00:00:00Z"],
            "uuids": [UUID1],
        }
    )
    collection.data.insert(
        {
            "text": "two",
            "int": 2,
            "float": 2.0,
            "bool": False,
            "date": "2021-01-02T00:00:00Z",
            "uuid": UUID2,
            "texts": ["two"],
            "ints": [2],
            "floats": [2.0],
            "bools": [False],
            "dates": ["2021-01-02T00:00:00Z"],
            "uuids": [UUID2],
        }
    )

    res = collection.aggregate.over_all(
        filters=filter_,
        return_metrics=[Metrics("text").text(count=True, top_occurrences_value=True)],
    )
    assert isinstance(res.properties["text"], AggregateText)
    assert res.properties["text"].count == 1
    assert res.properties["text"].top_occurrences[0].value == "two"


def test_over_all_with_filters_ref(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[
            Property(name="text", data_type=DataType.TEXT),
        ]
    )
    collection.config.add_reference(
        ReferenceProperty(name="ref", target_collection=collection.name)
    )

    uuid1 = collection.data.insert({"text": "one"})
    collection.data.insert({"text": "two"}, references={"ref": uuid1})

    res = collection.aggregate.over_all(
        filters=Filter.by_ref_multi_target("ref", collection.name).by_property("text").equal("one"),
        return_metrics=[Metrics("text").text(count=True, top_occurrences_value=True)],
    )
    assert isinstance(res.properties["text"], AggregateText)
    assert res.properties["text"].count == 1
    assert res.properties["text"].top_occurrences[0].value == "two"

    query = lambda: collection.aggregate.over_all(  # noqa: E731
        filters=Filter.by_ref("ref").by_property("text").equal("one"),
        return_metrics=[Metrics("text").text(count=True, top_occurrences_value=True)],
    )
    if collection._connection._weaviate_version.is_lower_than(1, 29, 0):
        with pytest.raises(WeaviateInvalidInputError):
            query()
    else:
        res = query()
        assert isinstance(res.properties["text"], AggregateText)
        assert res.properties["text"].count == 1
        assert res.properties["text"].top_occurrences[0].value == "two"


def test_wrong_aggregation(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(properties=[Property(name="text", data_type=DataType.TEXT)])
    if collection._connection._weaviate_version.is_at_least(1, 29, 0):
        pytest.skip("GQL is only used for versions 1.28.4 and lower")
    with pytest.raises(WeaviateQueryError) as e:
        collection.aggregate.over_all(total_count=False)
    assert (
        e.value.message
        == "The query that you sent had no body so GraphQL was unable to parse it. You must provide at least one option to the aggregation method in order to build a valid query."
    )


@pytest.mark.parametrize(
    "option,expected_len",
    [
        ({"object_limit": 1}, 1),
        ({"certainty": 0.9}, 1),
        ({"distance": 0.1}, 1),
        ({"object_limit": 2}, 2),
        ({"certainty": 0.1}, 2),
        ({"distance": 0.9}, 2),
    ],
)
def test_near_object_aggregation(
    collection_factory: CollectionFactory, option: dict, expected_len: int
) -> None:
    collection = collection_factory(
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
            vectorize_collection_name=False
        ),
    )
    text_1 = "some text"
    text_2 = "nothing like the other one at all, not even a little bit"
    uuid = collection.data.insert({"text": text_1})
    collection.data.insert({"text": text_2})
    res: AggregateReturn = collection.aggregate.near_object(
        uuid,
        return_metrics=[
            Metrics("text").text(count=True, top_occurrences_count=True, top_occurrences_value=True)
        ],
        **option,
    )
    assert isinstance(res.properties["text"], AggregateText)
    assert res.properties["text"].count == expected_len
    assert len(res.properties["text"].top_occurrences) == expected_len
    assert text_1 in [
        top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
    ]
    if expected_len == 2:
        assert text_2 in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]
    else:
        assert text_2 not in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]


def test_near_object_missing_param(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
            vectorize_collection_name=False
        ),
    )
    text_1 = "some text"
    text_2 = "nothing like the other one at all, not even a little bit"
    uuid = collection.data.insert({"text": text_1})
    collection.data.insert({"text": text_2})
    with pytest.raises(WeaviateInvalidInputError) as e:
        collection.aggregate.near_object(
            uuid,
            return_metrics=[
                Metrics("text").text(
                    count=True, top_occurrences_count=True, top_occurrences_value=True
                )
            ],
        )
    assert (
        "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
        == e.value.message
    )


@pytest.mark.parametrize(
    "option,expected_len",
    [
        ({"object_limit": 1}, 1),
        ({"object_limit": 2}, 2),
    ],
)
def test_hybrid_aggregation(
    collection_factory: CollectionFactory, option: dict, expected_len: int
) -> None:
    collection = collection_factory(
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
            vectorize_collection_name=False
        ),
    )
    text_1 = "some text"
    text_2 = "nothing like the other one at all, not even a little bit"
    uuid = collection.data.insert({"text": text_1})
    obj = collection.query.fetch_object_by_id(uuid, include_vector=True)
    assert "default" in obj.vector
    collection.data.insert({"text": text_2})
    res: AggregateReturn = collection.aggregate.hybrid(
        None,
        alpha=1,
        vector=obj.vector["default"],
        return_metrics=[
            Metrics("text").text(count=True, top_occurrences_count=True, top_occurrences_value=True)
        ],
        **option,
    )
    assert isinstance(res.properties["text"], AggregateText)
    assert res.properties["text"].count == expected_len
    assert len(res.properties["text"].top_occurrences) == expected_len
    assert text_1 in [
        top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
    ]
    if expected_len == 2:
        assert text_2 in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]
    else:
        assert text_2 not in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]


@pytest.mark.parametrize("group_by", ["text", GroupByAggregate(prop="text", limit=1)])
def test_hybrid_aggregation_group_by(
    collection_factory: CollectionFactory, group_by: Union[str, GroupByAggregate]
) -> None:
    collection = collection_factory(
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
            vectorize_collection_name=False
        ),
    )

    text_1 = "some text"
    text_2 = "nothing like the other one at all, not even a little bit"
    collection.data.insert({"text": text_1})
    collection.data.insert({"text": text_2})

    querier = lambda: collection.aggregate.hybrid(  # noqa: E731
        "text",
        alpha=0,
        query_properties=["text"],
        group_by=group_by,
        total_count=True,
        object_limit=2,  # has no effect due to alpha=0
    )
    if collection._connection._weaviate_version.is_lower_than(1, 25, 0):
        with pytest.raises(WeaviateUnsupportedFeatureError):
            querier()
        return

    res = querier()
    assert res.groups[0].grouped_by.prop == "text"
    assert res.groups[0].grouped_by.value == "some text"
    assert res.groups[0].total_count == 1


@pytest.mark.parametrize("group_by", ["text", GroupByAggregate(prop="text", limit=1)])
def test_hybrid_aggregation_group_by_with_named_vectors(
    collection_factory: CollectionFactory, group_by: Union[str, GroupByAggregate]
) -> None:
    collection = collection_factory(
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=[
            Configure.NamedVectors.text2vec_contextionary(
                name="all", vectorize_collection_name=False
            )
        ],
    )

    text_1 = "some text"
    text_2 = "nothing like the other one at all, not even a little bit"
    collection.data.insert({"text": text_1})
    collection.data.insert({"text": text_2})

    querier = lambda: collection.aggregate.hybrid(  # noqa: E731
        "text",
        alpha=0,
        query_properties=["text"],
        group_by=group_by,
        total_count=True,
        object_limit=2,  # has no effect due to alpha=0
        target_vector="all",
    )
    if collection._connection._weaviate_version.is_lower_than(1, 25, 0):
        with pytest.raises(WeaviateUnsupportedFeatureError):
            querier()
        return

    res = querier()
    assert res.groups[0].grouped_by.prop == "text"
    assert res.groups[0].grouped_by.value == "some text"
    assert res.groups[0].total_count == 1


@pytest.mark.parametrize(
    "option,expected_len",
    [
        ({"object_limit": 1}, 1),
        ({"certainty": 0.9}, 1),
        ({"distance": 0.1}, 1),
        ({"object_limit": 2}, 2),
        ({"certainty": 0.1}, 2),
        ({"distance": 0.9}, 2),
    ],
)
def test_near_vector_aggregation(
    collection_factory: CollectionFactory, option: dict, expected_len: int
) -> None:
    collection_maker = lambda: collection_factory(  # noqa: E731
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
            vectorize_collection_name=False
        ),
    )

    collection = collection_maker()
    text_1 = "some text"
    text_2 = "nothing like the other one at all, not even a little bit"
    uuid = collection.data.insert({"text": text_1})
    obj = collection.query.fetch_object_by_id(uuid, include_vector=True)
    assert "default" in obj.vector
    collection.data.insert({"text": text_2})
    res: AggregateReturn = collection.aggregate.near_vector(
        obj.vector["default"],
        return_metrics=[
            Metrics("text").text(count=True, top_occurrences_count=True, top_occurrences_value=True)
        ],
        **option,
    )
    assert isinstance(res.properties["text"], AggregateText)
    assert res.properties["text"].count == expected_len
    assert len(res.properties["text"].top_occurrences) == expected_len
    assert text_1 in [
        top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
    ]
    if expected_len == 2:
        assert text_2 in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]
    else:
        assert text_2 not in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]


def test_near_vector_missing_param(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
            vectorize_collection_name=False
        ),
    )
    uuid_ = collection.data.insert({"text": "some text"})
    obj = collection.query.fetch_object_by_id(uuid_, include_vector=True)
    assert "default" in obj.vector
    with pytest.raises(WeaviateInvalidInputError) as e:
        collection.aggregate.near_vector(
            obj.vector["default"],
            return_metrics=[
                Metrics("text").text(
                    count=True, top_occurrences_count=True, top_occurrences_value=True
                )
            ],
        )
    assert (
        "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
        == e.value.message
    )


@pytest.mark.parametrize(
    "option,expected_len",
    [
        ({"object_limit": 1}, 1),
        ({"certainty": 0.9}, 1),
        ({"distance": 0.1}, 1),
        ({"object_limit": 2}, 2),
        ({"certainty": 0.1}, 2),
        ({"distance": 0.9}, 2),
        ({"move_away": Move(concepts="something", force=0.000001), "distance": 0.9}, 2),
        ({"move_away": Move(objects=UUID1, force=0.000001), "distance": 0.9}, 2),
        ({"move_away": Move(concepts=["something", "else"], force=0.000001), "distance": 0.9}, 2),
        ({"move_to": Move(objects=[UUID1, UUID2], force=0.000001), "distance": 0.9}, 2),
    ],
)
def test_near_text_aggregation(
    collection_factory: CollectionFactory, option: dict, expected_len: int
) -> None:
    collection = collection_factory(
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
            vectorize_collection_name=False
        ),
    )
    text_1 = "some text"
    text_2 = "nothing like the other one at all, not even a little bit"
    collection.data.insert({"text": text_1}, uuid=UUID1)
    collection.data.insert({"text": text_2}, uuid=UUID2)
    res: AggregateReturn = collection.aggregate.near_text(
        text_1,
        return_metrics=[
            Metrics("text").text(count=True, top_occurrences_count=True, top_occurrences_value=True)
        ],
        **option,
    )
    assert isinstance(res.properties["text"], AggregateText)
    assert res.properties["text"].count == expected_len
    assert len(res.properties["text"].top_occurrences) == expected_len
    assert text_1 in [
        top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
    ]
    if expected_len == 2:
        assert text_2 in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]
    else:
        assert text_2 not in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]


def test_near_text_missing_param(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
            vectorize_collection_name=False
        ),
    )
    text_1 = "some text"
    collection.data.insert({"text": text_1})
    with pytest.raises(WeaviateInvalidInputError) as e:
        collection.aggregate.near_text(
            text_1,
            return_metrics=[
                Metrics("text").text(
                    count=True, top_occurrences_count=True, top_occurrences_value=True
                )
            ],
        )
    assert (
        "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
        == e.value.message
    )


@pytest.mark.skip(
    "img2vec-neural removed from CI. This test should go in the e2e-tests repo instead"
)
@pytest.mark.parametrize("option", [{"object_limit": 1}, {"certainty": 0.9}, {"distance": 0.1}])
def test_near_image_aggregation(collection_factory: CollectionFactory, option: dict) -> None:
    collection = collection_factory(
        properties=[
            Property(name="rating", data_type=DataType.INT),
            Property(name="image", data_type=DataType.BLOB),
        ],
        vectorizer_config=Configure.Vectorizer.img2vec_neural(image_fields=["image"]),
    )
    img_path = pathlib.Path("integration/weaviate-logo.png")
    collection.data.insert({"image": file_encoder_b64(img_path), "rating": 9})
    res: AggregateReturn = collection.aggregate.near_image(
        img_path,
        return_metrics=[Metrics("rating").integer(count=True, maximum=True)],
        **option,
    )
    assert isinstance(res.properties["rating"], AggregateInteger)
    assert res.properties["rating"].count == 1
    assert res.properties["rating"].maximum == 9


@pytest.mark.skip(
    "img2vec-neural removed from CI. This test should go in the e2e-tests repo instead"
)
def test_near_image_missing_param(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[
            Property(name="rating", data_type=DataType.INT),
            Property(name="image", data_type=DataType.BLOB),
        ],
        vectorizer_config=Configure.Vectorizer.img2vec_neural(image_fields=["image"]),
    )
    img_path = pathlib.Path("integration/weaviate-logo.png")
    collection.data.insert({"image": file_encoder_b64(img_path), "rating": 9})
    with pytest.raises(WeaviateInvalidInputError) as e:
        collection.aggregate.near_image(
            img_path,
            return_metrics=[
                Metrics("text").text(
                    count=True, top_occurrences_count=True, top_occurrences_value=True
                )
            ],
        )
    assert (
        "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
        == e.value.message
    )


def test_group_by_aggregation_argument(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[
            Property(name="text", data_type=DataType.TEXT),
            Property(name="int", data_type=DataType.INT),
        ],
    )
    collection.data.insert({"text": "some text", "int": 1})
    collection.data.insert({"text": "some text", "int": 2})

    res = collection.aggregate.over_all(
        group_by="text",
        return_metrics=[
            Metrics("text").text(count=True),
            Metrics("int").integer(count=True),
        ],
    )
    groups = res.groups
    assert len(groups) == 1
    assert groups[0].grouped_by.prop == "text"
    assert groups[0].grouped_by.value == "some text"
    assert isinstance(groups[0].properties["text"], AggregateText)
    assert groups[0].properties["text"].count == 2
    assert isinstance(groups[0].properties["int"], AggregateInteger)
    assert groups[0].properties["int"].count == 2

    res = collection.aggregate.over_all(
        group_by="int",
        return_metrics=[
            Metrics("text").text(count=True),
            Metrics("int").integer(count=True),
        ],
    )
    groups = res.groups
    assert len(groups) == 2
    assert groups[0].grouped_by.prop == "int"
    if collection._connection._weaviate_version.is_lower_than(1, 29, 0):
        assert groups[0].grouped_by.value == "1" or groups[1].grouped_by.value == "1"
    else:
        assert groups[0].grouped_by.value == 1 or groups[1].grouped_by.value == 1
    assert isinstance(groups[0].properties["text"], AggregateText)
    assert groups[0].properties["text"].count == 1
    assert isinstance(groups[0].properties["int"], AggregateInteger)
    assert groups[0].properties["int"].count == 1
    assert groups[1].grouped_by.prop == "int"
    if collection._connection._weaviate_version.is_lower_than(1, 29, 0):
        assert groups[1].grouped_by.value == "2" or groups[0].grouped_by.value == "2"
    else:
        assert groups[1].grouped_by.value == 2 or groups[0].grouped_by.value == 2
    assert isinstance(groups[1].properties["text"], AggregateText)
    assert groups[1].properties["text"].count == 1
    assert isinstance(groups[1].properties["int"], AggregateInteger)
    assert groups[1].properties["int"].count == 1


@pytest.mark.skip(reason="Validation logic is not robust enough currently")
def test_mistake_in_usage(
    collection_factory_get: CollectionFactoryGet, request: SubRequest
) -> None:
    collection = collection_factory_get(request.node.name)
    with pytest.raises(TypeError) as e:
        collection.aggregate.over_all([Metrics("text")])  # type: ignore # testing incorrect usage
    assert (
        e.value.args[0]
        == "One of the aggregations is an unexpected type: <class 'weaviate.collection.classes.aggregate.Metrics'>. Did you forget to append a method call? E.g. .text(count=True)"
    )
    with pytest.raises(TypeError) as e:
        collection.aggregate.over_all(aggregations=[Metrics("text")])  # type: ignore # testing incorrect usage
    assert (
        e.value.args[0]
        == "One of the aggregations is an unexpected type: <class 'weaviate.collection.classes.aggregate.Metrics'>. Did you forget to append a method call?  E.g. .text(count=True)"
    )


def test_all_available_aggregations(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[
            Property(name="text", data_type=DataType.TEXT),
            Property(
                name="texts",
                data_type=DataType.TEXT_ARRAY,
            ),
            Property(name="int", data_type=DataType.INT),
            Property(name="ints", data_type=DataType.INT_ARRAY),
            Property(name="float", data_type=DataType.NUMBER),
            Property(name="floats", data_type=DataType.NUMBER_ARRAY),
            Property(name="bool", data_type=DataType.BOOL),
            Property(name="bools", data_type=DataType.BOOL_ARRAY),
            Property(name="date", data_type=DataType.DATE),
            Property(name="dates", data_type=DataType.DATE_ARRAY),
        ],
    )
    collection.config.add_reference(
        ReferenceProperty(name="ref", target_collection=collection.name)
    )
    collection.data.insert(
        {
            "text": "some text",
            "texts": ["some text", "some text"],
            "int": 1,
            "ints": [1, 2],
            "float": 1.0,
            "floats": [1.0, 2.0],
            "bool": True,
            "bools": [True, False],
            "date": "2021-01-01T00:00:00Z",
            "dates": ["2021-01-01T00:00:00Z", "2021-01-02T00:00:00Z"],
        }
    )
    res = collection.aggregate.over_all(
        return_metrics=[
            Metrics("text").text(),
            Metrics("texts").text(),
            Metrics("int").integer(),
            Metrics("ints").integer(),
            Metrics("float").number(),
            Metrics("floats").number(),
            Metrics("bool").boolean(),
            Metrics("bools").boolean(),
            Metrics("date").date_(),
            Metrics("dates").date_(),
        ]
    )

    text = res.properties["text"]
    assert isinstance(text, AggregateText)
    assert text.count == 1
    assert text.top_occurrences[0].count == 1
    assert text.top_occurrences[0].value == "some text"

    texts = res.properties["texts"]
    assert isinstance(texts, AggregateText)
    assert texts.count == 2
    assert texts.top_occurrences[0].count == 2
    assert texts.top_occurrences[0].value == "some text"

    int_ = res.properties["int"]
    assert isinstance(int_, AggregateInteger)
    assert int_.count == 1
    assert int_.maximum == 1
    assert int_.mean == 1
    assert int_.median == 1
    assert int_.minimum == 1
    assert int_.mode == 1
    assert int_.sum_ == 1

    ints = res.properties["ints"]
    assert isinstance(ints, AggregateInteger)
    assert ints.count == 2
    assert ints.maximum == 2
    assert ints.mean == 1.5
    assert ints.median == 1.5
    assert ints.minimum == 1
    assert ints.mode == 1

    float_ = res.properties["float"]
    assert isinstance(float_, AggregateNumber)
    assert float_.count == 1
    assert float_.maximum == 1.0
    assert float_.mean == 1.0
    assert float_.median == 1.0
    assert float_.minimum == 1.0
    assert float_.mode == 1.0

    floats = res.properties["floats"]
    assert isinstance(floats, AggregateNumber)
    assert floats.count == 2
    assert floats.maximum == 2.0
    assert floats.mean == 1.5
    assert floats.median == 1.5
    assert floats.minimum == 1.0
    assert floats.mode == 1.0

    bool_ = res.properties["bool"]
    assert isinstance(bool_, AggregateBoolean)
    assert bool_.count == 1
    assert bool_.percentage_false == 0
    assert bool_.percentage_true == 1
    assert bool_.total_false == 0
    assert bool_.total_true == 1

    bools = res.properties["bools"]
    assert isinstance(bools, AggregateBoolean)
    assert bools.count == 2
    assert bools.percentage_false == 0.5
    assert bools.percentage_true == 0.5
    assert bools.total_false == 1
    assert bools.total_true == 1

    date = res.properties["date"]
    assert isinstance(date, AggregateDate)
    assert date.count == 1
    assert date.maximum == "2021-01-01T00:00:00Z"
    assert date.median == "2021-01-01T00:00:00Z"
    assert date.minimum == "2021-01-01T00:00:00Z"
    assert date.mode == "2021-01-01T00:00:00Z"

    dates = res.properties["dates"]
    assert isinstance(dates, AggregateDate)
    assert dates.count == 2
    assert dates.maximum == "2021-01-02T00:00:00Z"
    assert dates.median == "2021-01-01T12:00:00Z"
    assert dates.minimum == "2021-01-01T00:00:00Z"
    # assert res.properties["dates"].mode == "2021-01-02T00:00:00Z" # flakey: sometimes return 01, other times 02


def test_hybrid_bm25_operators(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[Property(name="name", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    if collection._connection._weaviate_version.is_lower_than(1, 31, 0):
        pytest.skip("bm25 operators are only supported in versions higher than 1.31.0")

    collection.data.insert({"name": "banana one"})
    collection.data.insert({"name": "banana two"})
    collection.data.insert({"name": "banana three"})
    collection.data.insert({"name": "banana four"})

    res: AggregateReturn = collection.aggregate.hybrid(
        query="banana two",
        alpha=0,
        bm25_operator=wvc.query.BM25Operator.or_(minimum_match=1),
        object_limit=10,
    )
    assert res.total_count == 4



import weaviate
from weaviate.client import WeaviateClient
from weaviate.collections.classes.config import Configure, DataType, Property


def test_raw_gql_v4() -> None:
    name = "RawGraphQlTest"
    number = 20
    client: WeaviateClient = weaviate.connect_to_local()
    collection = client.collections.create(
        name=name,
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    try:
        collection.data.insert_many([{"Name": f"name {i}"} for i in range(number)])

        response = client.graphql_raw_query(
            """{
                                            Aggregate {RawGraphQlTest{meta {count}}}
                                            Get{RawGraphQlTest{name}}
                                            }"""
        )

        assert response.errors is None
        assert response.aggregate[name][0]["meta"]["count"] == number
        assert len(response.get[name]) == number
    finally:
        client.collections.delete(name)
        client.close()


def test_raw_gql_v4_error() -> None:
    with weaviate.connect_to_local() as client:
        try:
            client.collections.create(name="dummy")
            response = client.graphql_raw_query(
                """{
                                                Get{IDoNotExist{name}}
                                                }"""
            )

            assert response.errors is not None
            assert len(response.get) == 0
            assert len(response.aggregate) == 0
            assert len(response.explore) == 0
        finally:
            client.collections.delete("dummy")



from dataclasses import dataclass
from typing import Any, List, Sequence, Union, get_args, get_origin

from weaviate.exceptions import WeaviateInvalidInputError
from weaviate.str_enum import BaseEnum


@dataclass
class _ValidateArgument:
    expected: List[Any]
    name: str
    value: Any


class _ExtraTypes(str, BaseEnum):
    NUMPY = "numpy"
    PANDAS = "pandas"
    POLARS = "polars"
    TF = "tensorflow"


def _validate_input(inputs: Union[List[_ValidateArgument], _ValidateArgument]) -> None:
    """Validate the values of the input arguments in comparison to the expected types defined in _ValidateArgument.

    It is not completely robust so be careful supplying subscripted generics in expected as it may not function as expected.
    To avoid this, only supply simply generics like Sequence[...] and List[...] as seen below in __is_valid.
    """
    if isinstance(inputs, _ValidateArgument):
        inputs = [inputs]
    for validate in inputs:
        if not any(_is_valid(exp, validate.value) for exp in validate.expected):
            raise WeaviateInvalidInputError(
                f"Argument '{validate.name}' must be one of: {validate.expected}, but got {type(validate.value)}"
            )


def _is_valid(expected: Any, value: Any) -> bool:
    if expected is None:
        return value is None

    # check for types that are not installed
    # https://stackoverflow.com/questions/12569452/how-to-identify-numpy-types-in-python
    if isinstance(expected, _ExtraTypes):
        return expected.value in type(value).__module__

    expected_origin = get_origin(expected)
    if expected_origin is Union:
        args = get_args(expected)
        return any(isinstance(value, arg) for arg in args)
    if expected_origin is not None and (
        issubclass(expected_origin, Sequence) or expected_origin is list
    ):
        if not isinstance(value, Sequence) and not isinstance(value, list):
            return False
        args = get_args(expected)
        if len(args) == 1:
            if get_origin(args[0]) is Union:
                union_args = get_args(args[0])
                return any(isinstance(val, union_arg) for val in value for union_arg in union_args)
            else:
                return all(isinstance(val, args[0]) for val in value)
    return isinstance(value, expected)
